import networkx as nx
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

from warnings import warn

# from cereeberus.compute.merge import isMerge
# from cereeberus.compute.merge import computeMergeTree

from .reebgraph import ReebGraph


class MergeTree(ReebGraph):
    r"""
    A merge tree stored as a ``ReebGraph`` object. Like a Reeb graph, this is a directed graph with a function defined on the vertices. However, in a merge tree, the function is required to iave a single root with function value treated as :math:`\infty`.

    We also store label information to construct a labeled merge tree. Here, this is a dictionary from some set (usually [1,...,n]) to a subset of vertices of the graph.

    """

    def __init__(self, T=None, root=None, f={}, labels={}, seed=None, verbose=False):
        """
        Initialize a merge tree object.

        Parameters:
            T: nx.graph, optional. If not none, it should be a tree with a specified root and function values.
            labels: dict, optional. A dictionary from labels to subsets of vertices.

        """

        if T == None:
            super().__init__(seed=seed, verbose=verbose)

            # Always add the infinite node
            self.add_node("v_inf", np.inf)
        else:
            # Check if the input is valid and add it
            T, f = self.inputToDirRootTree(T, root, f)
            super().__init__(T, f, seed, verbose)

            # Fix up the drawing locations
            self.fix_pos_f()

        self.labels = labels

    def __str__(self):
        return f"MergeTree with {len(self.nodes)} nodes and {len(self.edges)} edges."

    # ------------------------------------------------------#
    # Methods for listing, adding and removing nodes and edges
    # ---------------------------------------------------------#

    def get_finite_nodes(self):
        """
        Returns a list of the finite nodes of the tree. That is, everything except for `v_inf`.
        """
        return [v for v in self.nodes if v != "v_inf"]

    def get_leaves(self):
        """
        Returns a list of the leaves of the tree.
        """

        return [v for v in self.nodes if self.down_degree(v) == 0]

    def add_node(self, vertex, f_vertex, reset_pos=True):
        """
        Adds a node to the tree. Note that this will break the single connected component property of the tree so we assume you will do this in the process of adding more connecting edges.
        """

        super().add_node(vertex, f_vertex, reset_pos)
        warn(
            "Merge trees assume a single connected component. Adding a node may break this assumption."
        )

    def remove_node(self, vertex, reset_pos=True):
        """
        Removes a node from the tree. Note that this will break the single connected component property of the tree if it is not a leaf so we assume you will do this in the process of adding more connecting edges.
        """
        if vertex == "v_inf":
            raise ValueError("You cannot remove the infinite node")

        if self.down_degree(vertex) > 0:
            warn(
                "You removing a vertex with multiple children will create multiple connected components, so ensure that this property is eventually maintained."
            )

        super().remove_node(vertex, reset_pos)

    def add_edge(self, u, v, reset_pos=True):
        """
        Adds the edge (u,v) to the tree. This command will not allow you to add an edge if it will create a loop.
        """

        if self.f[u] < self.f[v]:
            lowerVertex = u
        else:
            lowerVertex = v

        if self.up_degree(lowerVertex) > 0:
            raise ValueError(
                f"Edge ({u,v}) cannot be added. Adding the edge will create a loop in the graph."
            )

        super().add_edge(u, v, reset_pos)

    def inputToDirRootTree(self, T, root, f):
        """
        Convert the input to a directed rooted tree that respects the internal structure.

        If T is undirected, it will be converted to a directed tree with the root specified.

        If f is not specified, a function will be induced from the tree from number of edges from the root.

        A vertex will be added with funciton value infinity.
        """
        if T == None:
            # Returns Tree, function
            return None, None

        elif nx.is_tree(T) == False:
            raise ValueError("The input graph is not a tree.")

        elif nx.is_directed(T) == False:
            if root == None:
                raise ValueError("You must specify a root for the tree.")
            else:
                # Convert to directed graph eminating from the root
                # Because of the ReebGraph convention, edges point up to higher function values.
                T = nx.bfs_tree(T, source=root).reverse()
        else:  # T is already directed
            # Make sure that this thing will be valid,
            # I think this is the same as "arboresence" in networkx
            # which is true if maximum in-degree equal to 1.
            # Since our graphs point upwards, this is the same as
            # making sure the reverse of our tree is an arboresence.

            if not nx.is_arborescence(T.reverse()):
                raise ValueError("The input tree is not a valid tree.")

        # No matter what, add a dummy node, point the old root at it and
        # eventually give that new node function value infinity.
        T.add_node("v_inf")
        T.add_edge(root, "v_inf")

        # Check that the function is valid
        if f == {}:
            # Define a function with distance from the root (but reversed). Rigged so the minimum is always at 0.
            f = nx.shortest_path_length(T, target=root)
            M = max(f.values())
            f = {i: M - f[i] for i in f.keys()}
            f["v_inf"] = np.inf

        else:
            for e in T.edges:
                if f[e[0]] < f[e[1]]:
                    raise ValueError(
                        "The function values should point upwards in the tree."
                    )

        return T, f

    # -----------------------------------------------#
    # Methods for drawing
    # -----------------------------------------------#

    def fix_pos_f(self):
        """
        Update drawing locations to deal with the fact that we have np.inf around.

        This sets the drawing location of the infinite node to be directly above the maximum finite node of the tree.
        """
        # Get the neighbor of the infinite node. Note we're assuming a single connected component so there's only one.
        top_vertices = list(self.predecessors("v_inf"))

        if len(top_vertices) == 0:
            warn("The infinite node has no neighbors so there is nothing to be done.")
        else:
            top_vertex = top_vertices[0]

            # Get  min and max function values
            Lmin = min(self.f.values())
            Lmax = self.f[top_vertex]
            height = Lmax - Lmin

            # Get the drawing location of the neighbor to have same x coordinate as the neighbor, but y coordinate at .3*height above the maximum function value.
            self.pos_f["v_inf"] = (self.pos_f[top_vertex][0], Lmax + 0.3 * height)

    def set_pos_from_f(self, seed=None, verbose=False):
        """
        Fix the drawing locations for the function values.
        """
        super().set_pos_from_f(seed, verbose)
        if "v_inf" in self.pos_f.keys():
            self.fix_pos_f()

    def draw(self, with_labels=True, label_type="names", with_colorbar=False):
        """
        Draw the merge tree.

        If `with_labels` is True, the labels will be drawn. This is either the vertex names if `label_type` is 'names' or the merge tree labels if `label_type` is 'labels'.
        """
        # viridis = mpl.colormaps['viridis'].resampled(16)
        fig, ax = plt.subplots()

        color_map = [self.pos_f[v][1] for v in self.nodes]

        nx.draw_networkx_nodes(self, self.pos_f, node_color=color_map)
        nx.draw_networkx_edges(self, self.pos_f)

        if with_labels:
            if label_type == "names":
                nx.draw_networkx_labels(self, self.pos_f)
            elif label_type == "labels":
                label_map = {}

                for k, v in self.labels.items():
                    label_map[v] = label_map.get(v, []) + [k]

                nx.draw_networkx_labels(self, self.pos_f, labels=label_map)
            else:
                raise ValueError("The label type must be either 'names' or 'labels'.")

        ax.tick_params(left=True, bottom=False, labelleft=True, labelbottom=False)

    # -----------------------------------------------#
    # Methods for labeled merge trees
    #
    # -----------------------------------------------#

    def label_all_leaves(self):
        """
        Label all the leaves of the tree by adding them to the labels dictionary if they're not already there.
        """
        leaves = self.get_leaves()

        if len(self.labels) == 0:
            key = 0
        else:
            key = max(self.labels.keys())
        for i, v in enumerate(leaves):
            if v not in self.labels.values():
                self.labels[key] = v
                key += 1

    def add_label(self, vertex, label=None):
        """
        Add a label to a vertex. If not provided, the label will be the next available integer.
        """
        if label == None:
            key = max(self.labels.keys()) + 1
        elif label in self.labels.keys():
            warn(f"The label {label} is already in use but will be reassigned.")
            key = label
        else:
            key = label

        self.labels[key] = vertex

    def add_label_edge(self, u, v, w, f_w, label=None):
        """
        Add a new vertex named `w` at function value `f_w` by subdividing the edge (u,v) and label it.

        """

        # Add the vertex
        self.subdivide_edge(u, v, w, f_w)

        # Add the label
        self.add_label(w, label)

    def LCA(self, u, v):
        """
        Compute the least common ancestor of two vertices in the tree.
        """
        # Get the path to the root from each
        anc_u = set(nx.shortest_path(self, u, "v_inf"))
        anc_v = set(nx.shortest_path(self, v, "v_inf"))

        # Get the intersection of the ancestors
        common_anc = anc_u.intersection(anc_v)

        # Get the vertex with the minimum function value
        return min(common_anc, key=lambda x: self.f[x])

    def LCA_matrix(self, type="leaves", return_as_df=False):
        """
        Compute the matrix of least common ancestors.

        If type is `leaves`, then the rows and columns of the matrix are determined by the leaf sets.

        If type is `labels`, then the rows and columns of the matrix are determined by the labels internal to the MergeTree.
        """
        if type == "leaves":
            # print('leaf version')
            nodes = self.get_leaves()
            # keys = list(range(len(nodes)))
            # print(nodes,keys)
            col_labels = nodes

        elif type == "labels":
            col_labels = list(self.labels.keys())
            nodes = [self.labels[i] for i in col_labels]

        else:
            raise ValueError(
                "The input `type` of LCA matrix must be either 'leaves' or 'labels'."
            )

        # Initialize the matrix
        n = len(nodes)
        M = np.zeros((n, n))

        # Compute the LCA matrix
        for i, node_i in enumerate(nodes):
            for j, node_j in enumerate(nodes):
                if i == j:
                    M[i, j] = self.f[node_i]
                else:
                    LCA_vertex = self.LCA(node_i, node_j)
                    M[i, j] = self.f[LCA_vertex]
        if return_as_df:
            return pd.DataFrame(M, col_labels, col_labels)
        else:
            return M


if __name__ == "__main__":
    from cereeberus.data.randomMergeTrees import randomMerge

    R = randomMerge(10)
    M = Merge(R.G, R.fx)
    M.plot_reeb()



from cereeberus import MapperGraph
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
import random


def set_global_seed(seed):
    """Set the global random seed.

    Parameters:
    seed (int): The seed to set.
    """

    global GLOBAL_SEED
    GLOBAL_SEED = seed
    random.seed(GLOBAL_SEED)


def generate_heights(length):
    """Generate a range of heights.

    Parameters:
    length (int): The number of heights to generate.

    Returns:
    list: A list of heights.
    """
    return range(1, length + 1)


def create_node_dict(width, heights):
    """
    Create a dictionary of nodes with their heights. Ensures at least one node exists at each height.

    Parameters:
    width (int): The maximum number of nodes that can exist at a height.
    heights (list): A list of heights.

    Returns:
    dict: A dictionary of nodes with their heights.

    """
    node_dict = {}
    for h in heights:
        num_nodes = random.randint(1, width)
        for i in range(num_nodes):
            node_dict[f"{h}_{i}"] = h
    return node_dict


def create_edge_list(node_dict, cutoff):
    """
    Create edges between nodes from adjacent heights. Edges are added based on the cutoff probability.

    Parameters:
    node_dict (dict): A dictionary of nodes with their heights.
    cutoff (float): The probability of adding an edge between nodes from adjacent heights.

    Returns:
    list: A list of edges.
    """
    height_dict = {}
    for node, height in node_dict.items():
        height_dict.setdefault(height, []).append(node)

    edge_list = []
    for height, nodes in height_dict.items():
        for adj_height in (height - 1, height + 1):
            if adj_height in height_dict:
                adj_nodes = height_dict[adj_height]
                for node1 in nodes:
                    edge_list.extend(
                        (node1, node2)
                        for node2 in adj_nodes
                        if random.random() <= cutoff
                    )
    return edge_list


def create_graph(width, length, cutoff, verbose=False, seed=9):
    """
    Create a connected NetworkX graph with nodes and edges based on parameters.

    Parameters:
    width (int): The maximum number of nodes that can exist at a height.
    length (int): The number of heights to generate.
    cutoff (float): The probability of adding an edge between nodes from adjacent heights.
    verbose (bool): Whether to print the nodes and edges of the graph.
    seed (int): The seed to set.

    Returns:
    nx.Graph: A connected NetworkX graph.
    """
    set_global_seed(seed)

    while True:
        heights = generate_heights(length)
        node_dict = create_node_dict(width, heights)
        edge_list = create_edge_list(node_dict, cutoff)

        G = nx.Graph()
        G.add_nodes_from(
            (node, {"func_val": value}) for node, value in node_dict.items()
        )
        G.add_edges_from(edge_list)

        # Exit the loop if the graph is connected
        if nx.is_connected(G):
            break

    if verbose:
        pos = {node: (int(node.split("_")[1]), node_dict[node]) for node in G.nodes()}
        print(f"Nodes: {G.nodes()}")
        print(f"Edges: {G.edges()}")
        nx.draw(G, pos, with_labels=True)

    return G, node_dict


def create_mapper_graph(width, length, cutoff, verbose=False, seed=9):
    """
    Create a MapperGraph object with nodes and edges based on parameters.

    Parameters:
    width (int): The maximum number of nodes that can exist at a height.
    length (int): The number of heights to generate.
    cutoff (float): The probability of adding an edge between nodes from adjacent heights.
    verbose (bool): Whether to print the nodes and edges of the graph.
    seed (int): The seed to set.

    Returns:
    MapperGraph: A MapperGraph object.
    """
    G, node_dict = create_graph(width, length, cutoff, verbose, seed)

    mapper_graph = MapperGraph(G, node_dict)
    return mapper_graph



import unittest
from cereeberus.distance.labeled_blocks import LabeledMatrix as LM 
from cereeberus.distance.labeled_blocks import LabeledBlockMatrix as LBM
import numpy as np

class TestBlockMatrixClasses(unittest.TestCase):

    def test_check_labeled_matrix(self):
        # 2 x 3
        Mat1 = LM([[0, 1, 2], [-3, -4, -5]], ['u', 'v'], ['a', 'b', 'c'])

        # 3 x 3
        Mat2 = LM([[0, 1], [1, 0], [0,0]], ['a', 'b', 'c'], ['x', 'y', 'z'])

        # 2 x 2
        Mat4 = LM([[0, 1], [1, 0]], ['a', 'b'], ['x', 'y'])

        # 3 x 3
        Mat5 =LM([[0, 1], [1, 0], [3,-2]], ['a', 'c', 'b'], ['x', 'y', 'z']) 

        # 2 x 2 
        Mat6 = LM([[0, 1], [1, 0]], ['u', 'v'], ['x', 'y'])
        Mat7 = LM([[0, 1], [1, 0]], ['a','b'], ['u', 'v'])

        # Check that the matrix is the right size
        self.assertEqual(Mat1.shape(), (2, 3))
        self.assertEqual(Mat1.size(), 6)
        self.assertEqual(Mat2.shape(), (3, 2))
        self.assertEqual(Mat2.size(), 6)

        # Check that we can multiply the matrices and the row and column labels are propogated correctly
        Mat3 = Mat1 @ Mat2
        self.assertEqual(Mat3.shape(), (Mat1.shape()[0], Mat2.shape()[1]))
        self.assertEqual(Mat3.rows, Mat1.rows)
        self.assertEqual(Mat3.cols, Mat2.cols)

        with self.assertRaises(ValueError):
            # Make sure an error is raised if we try to multiply two matrices of incompatible size. 
            Mat1 @ Mat4

            # Mat1 and Mat 5 have compatable sizes, but the row labels are not in the same order. This should raise an error.
            Mat1 @ Mat5

        # Check that we can ask for a matrix entry by label 
        self.assertEqual(Mat1['v', 'b'], -4)
        # and we can set a matrix entry by label
        Mat1['v', 'b'] = -100

        # We can add matrices if they have the same labeles 
        Mat1 + Mat1

        # We can't add matrices if they have different labels even if they have the same size 
        with self.assertRaises(ValueError):
            Mat1 + Mat6
            Mat1 + Mat7 

        # We have cases where we want to create a matrix that ends up being empty. 
        EmptyMat = LM(rows = ['a', 'b'], cols = [])
        self.assertEqual(EmptyMat.shape(), (2, 0))
        self.assertEqual(EmptyMat.size(), 0)
        EmptyMat2 = LM()
        self.assertEqual(EmptyMat2.shape(), (0, 0))
        self.assertEqual(EmptyMat2.size(), 0)

        # Check transpose function 
        Mat7 = Mat1.T()
        self.assertEqual(Mat7.shape(), (3, 2))
        self.assertEqual(Mat7.rows, Mat1.cols)
        self.assertEqual(Mat7.cols, Mat1.rows)

        # Column sum check 
        self.assertEqual(len(Mat1.col_sum()), Mat1.shape()[1])
        self.assertTrue(Mat2.is_col_sum_1())

        # Max and abs max check 
        self.assertEqual(Mat1.max(), 2)
        self.assertEqual(Mat1.absmax(), 100)
        self.assertTrue(np.isnan(EmptyMat.max()))
        self.assertTrue(np.isnan(EmptyMat.max()))
        # print('got here')

    def test_block_labeled_matrix(self):
        # Make the example from the front of the documentation 
        cols_dict = {1: ['a','b'], 2: ['c','d','e']}
        rows_dict = {0: ['u'], 1: ['v','w'], 2: ['x']}
        map_dict = {'a': 'v', 'b':'v', 'c': 'x', 'd':'x', 'e':'x'}
        lbm = LBM(map_dict, rows_dict, cols_dict)

        # A second example with rows and columns switched
        cols_dict = {0: ['u'], 1: ['v','w'], 2: ['x']}
        rows_dict = {0: ['x','y'], 1: ['z','a','b'], 2: ['c'], 3: ['d','e']}
        map_dict = {'u': 'y', 'v':'z', 'w': 'b', 'x':'c'}
        lbm2 = LBM(map_dict, rows_dict, cols_dict)


        # Check that the block matrix is the right size
        self.assertEqual(lbm.shape(), (4,5))

        # Check that the max and min functions work correctly 
        self.assertEqual(lbm.max(), 1)
        self.assertEqual(lbm.min(), 0)

        # Make sure we can take a transpose 
        lbm2T = lbm2.T()

        # Make sure we can add 
        lbm + lbm

        # Make sure we can't add if the labels are different
        with self.assertRaises(ValueError):
            lbm + lbm2

        # Make sure we can multiply, and that nothing freaks out if we have 
        # both a block matrix and a labeled matrix in the multiplication 
        lbm @ lbm.T() # Both block matrices

        bm = lbm.to_labeled_matrix()

        bm @ lbm.T() # Block matrix @ labeled matrix

        lbm @ bm.T() # Labeled matrix @ block matrix



import networkx as nx
import numpy as np
import matplotlib.pyplot as plt
from .reebgraph import ReebGraph


class EmbeddedGraph(nx.Graph):
    """
    A class to represent a graph with 2D embedded coordinates for each vertex.

    Attributes
        graph : nx.Graph
            a NetworkX graph object
        coordinates : dict
            a dictionary mapping vertices to their (x, y) coordinates

    """

    def __init__(self):
        """
        Initializes an empty EmbeddedGraph object.

        """
        super().__init__()
        self.coordinates = {}

    def add_node(self, vertex, x, y):
        """
        Adds a vertex to the graph and assigns it the given coordinates.

        Parameters:
            vertex (str):
                The vertex to be added.
            x (float):
                The x-coordinate of the vertex.
            y (float):
                The y-coordinate of the vertex.

        """
        super().add_node(vertex)
        self.coordinates[vertex] = (x, y)

    def add_nodes_from(self, nodes, coordinates):
        """
        Adds multiple vertices to the graph and assigns them the given coordinates.

        Parameters:
            nodes (list):
                A list of vertices to be added.
            coordinates (dict):
                A dictionary mapping vertices to their coordinates.

        """
        super().add_nodes_from(nodes)
        self.coordinates.update(coordinates)

    def add_edge(self, u, v):
        """
        Adds an edge between the vertices u and v if they exist.

        Parameters:
            u (str):
                The first vertex of the edge.
            v (str):
                The second vertex of the edge.

        """
        if not self.has_node(u) or not self.has_node(v):
            raise ValueError("One or both vertices do not exist in the graph.")
        else:
            super().add_edge(u, v)

    def get_coordinates(self, vertex):
        """
        Returns the coordinates of the given vertex.

        Parameters:
            vertex (str):
                The vertex whose coordinates are to be returned.

        Returns:
            tuple: The coordinates of the vertex.

        """
        return self.coordinates.get(vertex)

    def set_coordinates(self, vertex, x, y):
        """
        Sets the coordinates of the given vertex.

        Parameters:
            vertex (str):
                The vertex whose coordinates are to be set.
            x (float):
                The new x-coordinate of the vertex.
            y (float):
                The new y-coordinate of the vertex.

        Raises:
            ValueError: If the vertex does not exist in the graph.

        """
        if vertex in self.coordinates:
            self.coordinates[vertex] = (x, y)
        else:
            raise ValueError("Vertex does not exist in the graph.")

    def get_bounding_box(self):
        """
        Method to find a bounding box of the vertex coordinates in the graph.

        Returns:
            list: A list of tuples representing the minimum and maximum x and y coordinates.

        """
        if not self.coordinates:
            return None

        x_coords, y_coords = zip(*self.coordinates.values())
        return [(min(x_coords), max(x_coords)), (min(y_coords), max(y_coords))]

    def get_bounding_radius(self):
        """
        Method to find the radius of the bounding circle of the vertex coordinates in the graph.

        Returns:
            float: The radius of the bounding circle.

        """
        if not self.coordinates:
            return 0

        x_coords, y_coords = zip(*self.coordinates.values())
        norms = [np.linalg.norm(point) for point in zip(x_coords, y_coords)]

        return max(norms)

    def get_mean_centered_coordinates(self):
        """
        Method to find the mean-centered coordinates of the vertices in the graph.

        Returns:
            dict: A dictionary mapping vertices to their mean-centered coordinates.

        """
        if not self.coordinates:
            return None

        x_coords, y_coords = zip(*self.coordinates.values())
        mean_x, mean_y = np.mean(x_coords), np.mean(y_coords)

        return {v: (x - mean_x, y - mean_y) for v, (x, y) in self.coordinates.items()}

    def set_mean_centered_coordinates(self):
        """
        Method to set the mean-centered coordinates of the vertices in the graph. Warning: This overwrites the original coordinates

        """

        self.coordinates = self.get_mean_centered_coordinates()

    def g_omega(self, theta):
        """
        Function to compute the function :math:`g_\omega(v)` for all vertices :math:`v` in the graph in the direction of :math:`\\theta \in [0,2\pi]` . This function is defined by :math:`g_\omega(v) = \langle \\texttt{pos}(v), \omega \\rangle` .

        Parameters:

            theta (float):
                The angle in :math:`[0,2\pi]` for the direction to compute the :math:`g(v)` values.

        Returns:

            dict: A dictionary mapping vertices to their :math:`g(v)` values.

        """

        omega = (np.cos(theta), np.sin(theta))

        g = {}
        for v in self.nodes:
            g[v] = np.dot(self.coordinates[v], omega)
        return g

    def g_omega_edges(self, theta):
        """
        Calculates the function value of the edges of the graph by making the value equal to the max vertex value

        Parameters:

            theta (float):
                The direction of the function to be calculated.

        Returns:
            dict
                A dictionary of the function values of the edges.
        """
        g = self.g_omega(theta)

        g_edges = {}
        for e in self.edges:
            g_edges[e] = max(g[e[0]], g[e[1]])

        return g_edges

    def sort_vertices(self, theta, return_g=False):
        """
        Function to sort the vertices of the graph according to the function g_omega(v) in the direction of theta \in [0,2*np.pi].

        Parameters:
            theta (float):
                The angle in [0,2*np.pi] for the direction to sort the vertices.
            return_g (bool):
                Whether to return the g(v) values along with the sorted vertices.

        Returns:
            list
                A list of vertices sorted in increasing order of the :math:`g(v)` values.
                If ``return_g`` is True, also returns the ``g`` dictionary with the function values ``g[vertex_name]=func_value``.

        """
        g = self.g_omega(theta)

        v_list = sorted(self.nodes, key=lambda v: g[v])

        if return_g:
            # g_sorted = [g[v] for v in v_list]
            return v_list, g
        else:
            return v_list

    def sort_edges(self, theta, return_g=False):
        """
        Function to sort the edges of the graph according to the function

        .. math ::

            g_\omega(e) = \max \{ g_\omega(v) \mid  v \in e \}

        in the direction of :math:`\\theta \in [0,2\pi]` .

        Parameters:
            theta (float):
                The angle in :math:`[0,2\pi]` for the direction to sort the edges.
            return_g (bool):
                Whether to return the :math:`g(v)` values along with the sorted edges.

        Returns:
            A list of edges sorted in increasing order of the :math:`g(v)` values.
            If ``return_g`` is True, also returns the ``g`` dictionary with the function values ``g[vertex_name]=func_value``.

        """
        g_e = self.g_omega_edges(theta)

        e_list = sorted(self.edges, key=lambda e: g_e[e])

        if return_g:
            # g_sorted = [g[v] for v in v_list]
            return e_list, g_e
        else:
            return e_list

    def lower_edges(self, v, omega):
        """
        Function to compute the number of lower edges of a vertex v for a specific direction (included by the use of sorted v_list).

        Parameters:
            v (str):
                The vertex to compute the number of lower edges for.
            omega (tuple):
                The direction vector to consider given as an angle in [0, 2pi].

        Returns:
            int: The number of lower edges of the vertex v.

        """
        L = [n for n in self.neighbors(v)]
        gv = np.dot(self.coordinates[v], omega)
        Lg = [np.dot(self.coordinates[v], omega) for v in L]
        return sum(n >= gv for n in Lg)  # includes possible duplicate counts

    def plot(self, bounding_circle=False, color_nodes_theta=None, ax=None, **kwargs):
        """
        Function to plot the graph with the embedded coordinates.

        If ``bounding_circle`` is True, a bounding circle is drawn around the graph.

        If ``color_nodes_theta`` is not None, it should be given as a theta in :math:`[0,2\pi]`. Then the nodes are colored according to the :math:`g(v)` values in the direction of theta.

        """
        if ax is None:
            fig, ax = plt.subplots()
            # print("making new figure")
        else:
            fig = ax.get_figure()

        pos = self.coordinates
        if color_nodes_theta == None:
            nx.draw(self, pos, with_labels=True, ax=ax, **kwargs)
        else:
            g = self.g_omega(color_nodes_theta)
            color_map = [g[v] for v in self.nodes]
            # Some weird plotting to make the colorbar work.
            pathcollection = nx.draw_networkx_nodes(
                self, pos, node_color=color_map, ax=ax
            )
            nx.draw_networkx_labels(self, pos=pos, font_color="black", ax=ax)
            nx.draw_networkx_edges(self, pos, ax=ax, width=1, **kwargs)
            fig.colorbar(pathcollection, ax=ax, **kwargs)

        plt.axis("on")
        ax.tick_params(left=True, bottom=True, labelleft=True, labelbottom=True)

        if bounding_circle:
            r = self.get_bounding_radius()
            ax = plt.gca()
            circle1 = plt.Circle((0, 0), r, fill=False, linestyle="--", color="r")
            ax.add_patch(circle1)
            plt.axis("square")

        return ax

    def reeb_graph_from_direction(self, theta):
        """
        Function to create a ReebGraph from a given direction theta.

        Parameters:
            theta (float):
                The direction in [0, 2pi] to calculate the ReebGraph.

        Returns:
            ReebGraph: The ReebGraph object created from the EmbeddedGraph.

        """

        sorted_verts, g_verts = self.sort_vertices(theta, return_g=True)
        g_vert_list = np.array([g_verts[v] for v in sorted_verts])

        # Get the locations of a new function value since those will collapse down to a single vertex in the Reeb graph
        new_val_locs = (
            np.where(1 - np.isclose(g_vert_list[:-1], g_vert_list[1:]))[0] + 1
        )
        new_val_locs = np.concatenate(
            [
                [
                    0,
                ],
                new_val_locs,
            ]
        )

        # This dictionary will be find_comp[vertex_name] = new_vertex_name
        find_comp = {}
        vert_list = []
        f_dict_new = {}

        for val_i, i in enumerate(new_val_locs):

            if val_i == len(new_val_locs) - 1:
                verts = sorted_verts[i:]
            else:
                verts = sorted_verts[i : new_val_locs[val_i + 1]]

            L = list(nx.connected_components(nx.induced_subgraph(self, verts)))
            for cc_list in L:
                new_vert = "-".join(cc_list)
                vert_list.append(new_vert)
                for v in cc_list:
                    find_comp[v] = new_vert
                f_dict_new[new_vert] = g_vert_list[i]

        R = ReebGraph()
        R.add_nodes_from(vert_list, f_dict_new)

        for e in self.edges:
            if find_comp[e[0]] != find_comp[e[1]]:
                R.add_edge(find_comp[e[0]], find_comp[e[1]])

        return R


if __name__ == "__main__":
    # Example usage of the EmbeddedGraph class

    # Create an instance of the EmbeddedGraph class
    graph = EmbeddedGraph()

    # Add vertices with their coordinates
    graph.add_node("A", 1, 2)
    graph.add_node("B", 3, 4)
    graph.add_node("C", 5, 6)

    # Add edges between vertices
    graph.add_edge("A", "B")
    graph.add_edge("B", "C")

    # Get coordinates of a vertex
    coords = graph.get_coordinates("A")
    print(f"Coordinates of A: {coords}")

    # Set new coordinates for a vertex
    graph.set_coordinates("A", 7, 8)
    coords = graph.get_coordinates("A")
    print(f"New coordinates of A: {coords}")

    # Get the bounding box of the vertex coordinates
    bbox = graph.get_bounding_box()
    print(f"Bounding box: {bbox}")



# %--- Example usage of the EmbeddedGraph class ---%

from cereeberus import EmbeddedGraph


def exampleEmbeddedGraph(mean_centered=True):
    """
    Function to create an example ``EmbeddedGraph`` object. Helpful for testing.

    Returns:
        EmbeddedGraph: An example ``EmbeddedGraph`` object.

    """
    graph = EmbeddedGraph()

    graph.add_node("A", 1, 2)
    graph.add_node("B", 3, 4)
    graph.add_node("C", 5, 7)
    graph.add_node("D", 3, 6)
    graph.add_node("E", 4, 3)
    graph.add_node("F", 4, 5)

    graph.add_edge("A", "B")
    graph.add_edge("B", "C")
    graph.add_edge("B", "D")
    graph.add_edge("B", "E")
    graph.add_edge("C", "D")
    graph.add_edge("E", "F")

    if mean_centered:
        graph.set_mean_centered_coordinates()

    return graph



from ..reeb.lowerstar import LowerStar
import matplotlib.pyplot as plt
import numpy as np


class Torus(LowerStar):
    """
    Class to create an example torus using the LowerStar.

    Inherits from:
        LowerStar: A lower star simplicial complex.

    Methods:
        __init__(): Initializes the torus by inserting simplices.
        get_vertex_integer(i, j): Returns the unique integer for the vertex at (i, j).
        get_vertex_coordinates(vertex_int): Returns the (i, j) coordinates for a given vertex integer.
        plot(): Plots a flat representation of the torus's 1-skeleton.

    Example:
        >>> torus = Torus()
        >>> torus.generate_grid(grid_size=5)
        >>> torus.assign_random_values(0, 10)
        >>> torus.plot()

    """

    def __init__(self):
        """Initialize class. Due to SimplexTree C++ weirdness, this does nothing but initialize the instance. Call T.generate_grid(5) to add the content."""

        super().__init__()

    def generate_grid(self, grid_size=5):
        """Initializes a simplicial complex representing a torus. The grid_size is used to generate a mesh with (n-1)x(n-1) vertices. Visually, this could be drawn as a grid of nxn vertices, but where the last row and the last column are matched up.

        Args:
            grid_size (int, optional): _description_. Defaults to 5, Minimum is 4
        """

        if grid_size < 3:
            raise ValueError("grid_size must be at least 4 to form a torus.")

        self.grid_size = grid_size
        n = grid_size

        # Add triangles (two per square, with periodic boundary)
        # Note that vertices and edges are added automatically as needed

        for i in range(n):
            for j in range(n):
                v0 = self.get_vertex_integer(i, j)
                v1 = self.get_vertex_integer((i + 1) % n, j)
                v2 = self.get_vertex_integer(i, (j + 1) % n)
                v3 = self.get_vertex_integer((i + 1) % n, (j + 1) % n)
                # Lower triangle
                self.insert([v0, v1, v2])
                # Upper triangle
                self.insert([v1, v2, v3])

    def get_vertex_integer(self, i, j):
        """Returns the unique integer corresponding to the vertex at position (i, j) in the grid.

        Args:
            i (int): Row index.
            j (int): Column index.

        Returns:
            int: Unique integer representing the vertex.
        """
        n = self.grid_size
        return int(i * n + j)

    def get_vertex_coordinates(self, vertex_int):
        """Returns the (i, j) coordinates of a vertex given its unique integer.

        Args:
            vertex_int (int): Unique integer representing the vertex.

        Returns:
            tuple: (i, j) coordinates of the vertex.
        """
        n = self.grid_size
        i = vertex_int // n
        j = vertex_int % n
        return (i, j)

    def assign_filtration(self, vertex, value):
        """Sets the filtration value for a given vertex. Vertex can be provided either as an integer, as a single integer in a list representing the integer label of the vertex, or as (i, j) coordinates for the grid structure of the torus.

        Args:
            vertex (Union[int, tuple]): Unique integer representing the vertex or (i, j) coordinates.
            value (float): Filtration value to set.
        """

        if isinstance(vertex, int):
            vertex = [vertex]
        elif len(vertex) == 1:
            pass
        elif len(vertex) == 2:
            vertex = [self.get_vertex_integer(*vertex)]
        else:
            raise ValueError(
                "vertex must be an integer, a single-element list, or a tuple of (i, j) coordinates."
            )

        super().assign_filtration(vertex, value)

    def draw(self, ax=None, cmap=plt.cm.viridis, **kwargs):
        """Gives a flat plot of the torus's 1-skeleton using matplotlib.

        Args:
            ax (matplotlib.axes.Axes, optional): Axes to plot on. If None, a new figure and axes are created. Defaults to None.
            cmap (matplotlib.colors.Colormap, optional): Colormap to use for vertex colors based on filtration values. Defaults to plt.cm.viridis.
            **kwargs: Additional keyword arguments passed to plt.scatter for vertex plotting.

        """

        n = self.grid_size
        if ax is None:
            fig, ax = plt.subplots()

        # Get a grid of vertex positions. Each row is (i,j) in order
        n = self.grid_size  # or set n = 5
        I, J = np.meshgrid(np.arange(n), np.arange(n), indexing="ij")
        positions = np.stack([I.ravel(), J.ravel()], axis=1)

        vmin = self.min_filtration()
        vmax = self.max_filtration()
        norm = plt.Normalize(vmin=vmin, vmax=vmax)
        colors = cmap(norm([self.filtration([v]) for v in range(n * n)]))

        # Plot vertices
        plt.scatter(
            positions[:, 0],
            positions[:, 1],
            c=colors,
            s=100,
            edgecolors="k",
            zorder=2,
            **kwargs,
        )
        # Plot the vertices on the other size, but greyed out to indicate periodicity
        for i in range(n):
            ax.plot(i, n, "ko", alpha=0.3)  # Top row
            ax.plot(n, i, "ko", alpha=0.3)  # Right column
        ax.plot(n, n, "ko", alpha=0.3)  # Top-right corner

        plt.colorbar(
            plt.cm.ScalarMappable(norm=norm, cmap=cmap), ax=ax, label="Filtration Value"
        )

        # Plot edges
        for simplex in self.get_skeleton(1):
            if len(simplex[0]) == 2:  # Only consider edges
                v0, v1 = simplex[0]
                i0, j0 = self.get_vertex_coordinates(v0)
                i1, j1 = self.get_vertex_coordinates(v1)

                # Handle periodic boundary conditions
                if abs(i1 - i0) > 1:
                    if i0 == 0:
                        i0 += n
                    else:
                        i1 += n
                if abs(j1 - j0) > 1:
                    if j0 == 0:
                        j0 += n
                    else:
                        j1 += n

                ax.plot(
                    [i0, i1], [j0, j1], "k-", zorder=1
                )  # 'v-' means violet color, solid line

        # Add grey edges on the right and top
        for i in range(n + 1):
            # Top row edges
            ax.plot([i, (i + 1) % n], [n, n], "k--", alpha=0.3, zorder=1)
            # Right column edges
            ax.plot([n, n], [i, (i + 1) % n], "k--", alpha=0.3, zorder=1)

        ax.set_aspect("equal")
        ax.axis("off")
        plt.title("Flat representation of the Torus 1-skeleton")
        plt.grid(True)
        plt.show()



import unittest
from cereeberus import ReebGraph, LowerStar, computeReeb
from cereeberus.data import Torus


class TestLowerStarClass(unittest.TestCase):
    
    def test_lower_star_assign_filtration(self):
        # Test that assigning filtration values to vertices correctly updates adjacent simplices.
        K = LowerStar()
        K.insert([0, 1, 2])
        K.insert([1, 3])
        K.insert([2, 3])
        
        K.assign_filtration(0, 0.0)
        self.assertEqual(K.filtration([0]), 0.0)
        self.assertEqual(K.filtration([0, 1]), 0.0)
        self.assertEqual(K.filtration([0, 2]), 0.0)
        self.assertEqual(K.filtration([0, 1, 2]), 0.0)
        
        K.assign_filtration(1, 3.0)
        self.assertEqual(K.filtration([1]), 3.0)
        self.assertEqual(K.filtration([1, 3]), 3.0)
        self.assertEqual(K.filtration([0, 1]), 3.0)  # Updated due to vertex 1
        self.assertEqual(K.filtration([0, 1, 2]), 3.0)  # Updated due to vertex 1
        
    def test_lower_star_sc_max_min_filtration(self):
        # Test that max and min filtration values are computed correctly.
        K = LowerStar()
        K.insert([0, 1])
        K.insert([1, 2])
        K.insert([2, 3])
        
        K.assign_filtration(0, 1.0)
        K.assign_filtration(1, 2.0)
        K.assign_filtration(2, 3.0)
        K.assign_filtration(3, 4.0)
        
        self.assertEqual(K.min_filtration(), 1.0)
        self.assertEqual(K.max_filtration(), 4.0)
        
    def test_computeReeb(self):
        # Test the computation of the Reeb graph from a Lower Star Simplicial Complex.
        K = LowerStar()
        K.insert([0, 1, 2])
        K.insert([1, 3])
        K.insert([2, 3])
        
        K.assign_filtration(0, 0.0)
        K.assign_filtration(1, 3.0)
        K.assign_filtration(2, 5.0)
        K.assign_filtration(3, 7.0)
        
        R = computeReeb(K)
        
        self.assertIsInstance(R, ReebGraph)
        self.assertGreater(len(R.nodes), 0)
        self.assertGreater(len(R.edges), 0)
        
    def test_torus_example_class(self):
        # Test the torus example from the documentation.
        T = Torus()
        T.generate_grid(grid_size = 4)
        T.assign_random_values(0,100, seed=1986)
        
        R = computeReeb(T)
        
        self.assertIsInstance(R, ReebGraph)
        self.assertGreater(len(R.nodes), 0)
        self.assertGreater(len(R.edges), 0)


import os
import sys

import networkx as nx
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from scipy.linalg import block_diag

from ..compute.unionfind import UnionFind
from ..compute.utils import HiddenPrints
from .ilp import solve_ilp, solve_ilp_dist
from .labeled_blocks import LabeledBlockMatrix as LBM
from .labeled_blocks import LabeledMatrix as LM


class Interleave:
    # Import MapperGraph inside methods as needed to avoid circular import
    """
    A class to compute the interleaving distance between two Mapper graphs, denoted :math:`F` and :math:`G`. The interleaving distance is a measure of how similar two Mapper graphs are, based on the induced maps between them.

    Once the ``Interleave.fit()`` command has been run, the optimal bound for the distance for the input mapper graphs is stored as ``Interleave.n``. The resulting maps can be found using the ``Interleave.phi()`` and ``Interleave.psi()`` functions. For more detailed structure, the full interleaving is stored using an `Assignment` class, documented below, and can be accessed using ``Interleave.assignment``.

    """

    def __init__(self, F, G):
        """
        Initialize the Interleave object.

        Parameters:
            F (MapperGraph):
                The first Mapper graph.
            G (MapperGraph):
                The second Mapper graph.
        """

        self.F = F
        self.G = G
        self.n = np.inf
        self.assignment = None
        

    
    def fit(self, pulp_solver = None, verbose= False, max_n_for_error = 100):
        """
        Finds the smallest feasible n using exponential + binary search and stores the optimal assignment in self.assignment and the n in self.n.
        
        Parameters:
            pulp_solver (pulp.LpSolver): 
                solver to use for ILP. If None, the default solver is used.
            verbose (bool, optional): 
                If True, print the progress of the optimization. Defaults to False.
            max_n_for_error (int, optional): 
                The maximum value of `n` to search for. If the interleaving distance is not found by this value, a ValueError is raised. Defaults to 100. ####NOTE: this can be replaced by the bounding box.
            
        Returns:
            int: smallest feasible n
        """
        # catch the results to avoid recomputation
        checked_results = {}


        # Step 0: Check for smallest possible n (n=0 when they both have same function ranges)

        min_n = max(abs(self.F.min_f()-self.G.min_f()), abs(self.F.max_f()-self.G.max_f()))


        if min_n not in checked_results:
            myAssgn = Assignment(self.F, self.G, n=min_n)
            prob_status = myAssgn.optimize(pulp_solver=pulp_solver)

            checked_results[min_n] = (prob_status, myAssgn)

        prob_status, myAssgn = checked_results[min_n]
        if verbose:
            print(f"\n-\nTrying n = {min_n}...")
            print(f"n = {min_n}, status = {prob_status}")

        
        
        if prob_status == True:
            self.n = min_n
            self.assignment = myAssgn
            return self.n
        
        # Step 1: Exponential search for first feasible n
        low, high  = min_n, min_n+1
        found_feasible_n = False

        while high <= max_n_for_error:
            if high not in checked_results:
                myAssgn = Assignment(self.F, self.G, n=high)
                prob_status = myAssgn.optimize(pulp_solver=pulp_solver)
                checked_results[high] = (prob_status, myAssgn)

            prob_status, myAssgn = checked_results[high]

            if verbose:
                print(f"\n-\nTrying n = {high}...")
                print(f"n = {high}, status = {prob_status}")
            
            

            if prob_status == True:
                found_feasible_n = True
                break
            low, high = high, high * 2 # double n
        else:
            if not found_feasible_n:
                raise ValueError(f"Interleaving distance not found for n <= {max_n_for_error}.")
            

        high = min(high, max_n_for_error)  # Clamp to max allowed

        # Step 2: Binary search for smallest feasible n in [low+1, high]
        best_n = high

        while low <= high:
            mid = (low + high) // 2
            if mid not in checked_results:
                myAssgn = Assignment(self.F, self.G, n=mid)
                prob_status = myAssgn.optimize(pulp_solver=pulp_solver)
                checked_results[mid] = (prob_status, myAssgn)

            prob_status, myAssgn = checked_results[mid]

            if verbose:
                print(f"\n-\nTrying n = {mid}...")
                print(f"n = {mid}, status = {prob_status}")

            
            if prob_status == True:
                best_n = mid
                high = mid - 1 # try smaller n

            else:
                low = mid+1  # update to last infeasible n

        # Step 3: Final validation
        self.n = best_n
        prob_status, myAssgn = checked_results[self.n]
        self.assignment = myAssgn

        if prob_status != True:
            raise ValueError(f"Final fit object is not an interleaving. Status = {prob_status} for n = {self.n}.")
        
        return self.n



    def dist_fit(self, pulp_solver = None, verbose= False, max_n_for_error = 100):
        """
        Compute the interleaving distance between the two Mapper graphs.

        Parameters:
            pulp_solver (pulp.LpSolver):
                The solver to use for the ILP optimization. If None, the default solver is used.
            verbose (bool, optional):
                If True, print the progress of the optimization. Defaults to False.
            max_n_for_error (int, optional):
                The maximum value of `n` to search for. If the interleaving distance is not found by this value, a ValueError is raised. Defaults to 100. ####NOTE: this can be replaced by the bounding box.

        Returns:
            Int:
                The interleaving distance, that is, the smallest n for which loss is zero.
        """

        # catch the results to avoid recomputation
        checked_results = {}

        # Step 0: Check for smallest possible n (n=0 when they both have same function ranges)
        min_n = max(abs(self.F.min_f()-self.G.min_f()), abs(self.F.max_f()-self.G.max_f())) # minimum possible n based on function ranges


        if min_n not in checked_results:
            myAssgn = Assignment(self.F, self.G, n = min_n)
            Loss = myAssgn.dist_optimize(pulp_solver = pulp_solver)
            checked_results[min_n] = (Loss, myAssgn)
        Loss, myAssgn = checked_results[min_n]

        if verbose:
            print(f"\n-\nTrying n = {min_n}...")
            print(f"n = {min_n}, Loss = {Loss}, distance_bound = {min_n + Loss}")

        # if loss is 0, we're done
        if Loss == 0:
            self.n = min_n
            self.assignment = myAssgn
            return self.n

        # step 1: exponential search for the upperbound
        low, high = min_n, min_n+1
        found_valid_n = False

        while high <= max_n_for_error:
            try:
                if high not in checked_results:
                    myAssgn = Assignment(self.F, self.G, n = high)
                    Loss = myAssgn.dist_optimize(pulp_solver = pulp_solver)
                    checked_results[high] = (Loss, myAssgn)

                Loss, myAssgn = checked_results[high]

                if verbose:
                    print(f"\n-\nTrying n = {high}...")
                    print(f"n = {high}, Loss = {Loss}, distance_bound = {high + Loss}")

                if Loss == 0:
                    found_valid_n = True
                    break
                low, high = high, high * 2
            except ValueError:  # infeasible assignment
                low, high = high, high * 2

        if not found_valid_n:
            raise ValueError(
                f"Interleaving distance not found for n <= {max_n_for_error}."
            )

        high = min(high, max_n_for_error)  # Clamp to max allowed
        
        # step 2: binary search for the optimal n
        # low = (high//2) + 1 if high > 1 else 1
        low = low #keep the last infeasible n as the lower bound
        best_n = high

        while low <= high:
            mid = (low + high) // 2
            try:
                if mid not in checked_results:
                    myAssgn = Assignment(self.F, self.G, n = mid)
                    Loss = myAssgn.dist_optimize(pulp_solver = pulp_solver)
                    checked_results[mid] = (Loss, myAssgn)

                Loss, myAssgn = checked_results[mid]

                if verbose:
                    print(f"\n-\nTrying n = {mid}...")
                    print(f"n = {mid}, Loss = {Loss}, distance_bound = {mid + Loss}")

                
                if Loss == 0:
                    best_n = mid
                    high = mid - 1 # decrease n to increase the loss. this tries to  go lower
                else:
                    bound = mid + Loss
                    if bound < best_bound:
                        best_bound = bound # to tighten the upper bound on the search space. this tries to go higher
                    low = mid + 1
            except ValueError: # infeasible assignment
                low = mid + 1  
        
        # validate the final solution
        self.n = best_n
        Loss, myAssgn = checked_results[self.n]
        if Loss != 0:
            raise ValueError(f"Unexpected non-zero loss (Loss={Loss}) for n={self.n}")
        self.assignment = myAssgn
        return self.n
    

    # def og_dist_fit(self, pulp_solver = None, verbose= False, max_n_for_error = 100):
    #     """
    #     Compute the interleaving distance between the two Mapper graphs.
        
    #     Parameters:
    #         pulp_solver (pulp.LpSolver): 
    #             The solver to use for the ILP optimization. If None, the default solver is used.
    #         verbose (bool, optional): 
    #             If True, print the progress of the optimization. Defaults to False.
    #         max_n_for_error (int, optional): 
    #             The maximum value of `n` to search for. If the interleaving distance is not found by this value, a ValueError is raised. Defaults to 100. ####NOTE: this can be replaced by the bounding box.
            
    #     Returns:
    #         Int:
    #             The interleaving distance, that is, the smallest n for which loss is zero.
    #     """

    #     # catch the results to avoid recomputation
    #     checked_results = {}

    #     # Step 0: Check for smallest possible n (n=0 when they both have same function ranges)
    #     min_n = max(abs(self.F.min_f()-self.G.min_f()), abs(self.F.max_f()-self.G.max_f())) # minimum possible n based on function ranges


    #     if min_n not in checked_results:
    #         myAssgn = Assignment(self.F, self.G, n = min_n)
    #         Loss = myAssgn.dist_optimize(pulp_solver = pulp_solver)
    #         checked_results[min_n] = (Loss, myAssgn)
    #     Loss, myAssgn = checked_results[min_n]

    #     if verbose:
    #         print(f"\n-\nTrying n = {min_n}...")
    #         print(f"n = {min_n}, Loss = {Loss}, distance_bound = {min_n + Loss}")

    #     # if loss is 0, we're done
    #     if Loss == 0:
    #         self.n = min_n
    #         self.assignment = myAssgn
    #         return self.n
        
    #     # step 1: exponential search for the upperbound
    #     low, high = min_n, min_n+1
    #     found_valid_n = False

    #     while high <= max_n_for_error:
    #         try:
    #             if high not in checked_results:
    #                 myAssgn = Assignment(self.F, self.G, n = high)
    #                 Loss = myAssgn.dist_optimize(pulp_solver = pulp_solver)
    #                 checked_results[high] = (Loss, myAssgn)

    #             Loss, myAssgn = checked_results[high]

    #             if verbose:
    #                 print(f"\n-\nTrying n = {high}...")
    #                 print(f"n = {high}, Loss = {Loss}, distance_bound = {high + Loss}")

    #             if  Loss == 0:
    #                 found_valid_n = True
    #                 break
    #             low, high = high, high*2
    #         except ValueError: # infeasible assignment
    #             low, high = high, high*2

    #     if not found_valid_n:
    #         raise ValueError(f"Interleaving distance not found for n <= {max_n_for_error}.")
        
    #     high = min(high, max_n_for_error)  # Clamp to max allowed
    #     # step 2: binary search for the optimal n
        
    #     # low = (high//2) + 1 if high > 1 else 1
    #     low = low
    #     best_n = high

    #     while low <= high:
    #         mid = (low + high) // 2 
    #         try:
    #             if mid not in checked_results:
    #                 myAssgn = Assignment(self.F, self.G, n = mid)
    #                 Loss = myAssgn.dist_optimize(pulp_solver = pulp_solver)
    #                 checked_results[mid] = (Loss, myAssgn)

    #             Loss, myAssgn = checked_results[mid]

    #             if verbose:
    #                 print(f"\n-\nTrying n = {mid}...")
    #                 print(f"n = {mid}, Loss = {Loss}, distance_bound = {mid + Loss}")
                
    #             if Loss == 0:
    #                 best_n = mid
    #                 high = mid - 1 # decrease n to increase the loss
    #             else:
    #                 low = mid + 1
    #         except ValueError: # infeasible assignment
    #             low = mid + 1  
        
    #     # validate the final solution
    #     self.n = best_n
    #     Loss, myAssgn = checked_results[self.n]
    #     if Loss != 0:
    #         raise ValueError(f"Unexpected non-zero loss (Loss={Loss}) for n={self.n}")
    #     self.assignment = myAssgn
    #     return self.n
            
    
            
    

    def phi(self, key="0", obj_type="V"):
        """
        Get the interleaving map :math:`\\varphi: F \\to G^n` if ``key == '0'`` or :math:`\\varphi_n: F^n \\to G^{2n}` if ``key == 'n'``.

        Parameters:
            key (str) :
                The key for the map. Either ``'0'`` or ``'n'``.
            obj_type (str) :
                The type of map. Either ``'V'`` or ``'E'``.

        Returns:
            LabeledBlockMatrix :
                The interleaving map.
        """
        if self.assignment is None:
            raise ValueError("You must call fit() before getting the interleaving map.")
        return self.assignment.phi(key, obj_type)

    def psi(self, key="0", obj_type="V"):
        """
        Get the interleaving map :math:`\\psi: G \\to F^n` if ``key == '0'`` or :math:`\\psi_n: G^n \\to F^{2n}` if ``key == 'n'``.

        Parameters:
            key (str) :
                The key for the map. Either ``'0'`` or ``'n'``.
            obj_type (str) :
                The type of map. Either ``'V'`` or ``'E'``.

        Returns:
            LabeledBlockMatrix :
                The interleaving map.
        """
        if self.assignment is None:
            raise ValueError("You must call fit() before getting the interleaving map.")
        return self.assignment.psi(key, obj_type)

    def get_interleaving_map(self, maptype="phi", key="0", obj_type="V"):
        """
        Get the relevant interleaving map. Helpful for iterating over options.

        Parameters:
            maptype (str) :
                The type of map. Either ``'phi'`` or ``'psi'``.
            key (str) :
                The key for the map. Either ``'0'`` or ``'n'``.
            obj_type (str) :
                The type of map. Either ``'V'`` or ``'E'``.

        Returns:
            LabeledBlockMatrix :
                The relevant interleaving map.
        """

        if self.assignment is None:
            raise ValueError("You must call fit() before getting the interleaving map.")
        return self.assignment.get_interleaving_map(maptype, key, obj_type)

    def draw_all_graphs(self, figsize=(15, 10), **kwargs):
        """Draw all the graphs stored in the Interleave object.

        Args:
            figsize (tuple, optional): Sets the size of the figure. Defaults to (15,10).

        Returns:
            tuple: The figure and axes objects.
        """

        if self.assignment is None:
            raise ValueError("You must call fit() before drawing the graphs.")
        return self.assignment.draw_all_graphs()

    def draw_all_phi(self, figsize=(15, 10), **kwargs):
        """Draw all the phi maps stored in the Interleave object.

        Args:
            figsize (tuple, optional): Sets the size of the figure. Defaults to (15,10).
            **kwargs: Additional keyword arguments to pass to the drawing function.

        Returns:
            tuple: The figure and axes objects.
        """

        if self.assignment is None:
            raise ValueError("You must call fit() before drawing the phi maps.")
        return self.assignment.draw_all_phi(figsize, **kwargs)

    def draw_all_psi(self, figsize=(15, 10), **kwargs):
        """Draw all the psi maps stored in the Interleave object.

        Args:
            figsize (tuple, optional): Sets the size of the figure. Defaults to (15,10).
            **kwargs: Additional keyword arguments to pass to the drawing function.

        Returns:
            tuple: The figure and axes objects.
        """

        if self.assignment is None:
            raise ValueError("You must call fit() before drawing the psi maps.")
        return self.assignment.draw_all_psi(figsize, **kwargs)

    # Disable Printing
    def _blockPrint():
        sys.stdout = open(os.devnull, "w")

    # Restore Printing
    def _enablePrint():
        sys.stdout = sys.__stdout__


# ============================
# Assignment Class
# ============================


class Assignment:
    """
    A class to determine the loss for a given assignment, and thus bound the interleaving distance between two Mapper graphs, denoted :math:`F` and :math:`G` throughout.

    We use keys ``['0', 'n', '2n']`` to denote the Mapper graphs :math:`F = F_0`, :math:`F_n`, and :math:`F_{2n}` and similarly for :math:`G`.

    Note that the difference in the ranges of the two Mapper graphs must be within ``'n'``.
    """

    def __init__(self, F, G, n=1, initialize_random_maps=False, seed=None):
        """
        Initialize the Interleave object.

        Parameters:
            F (MapperGraph):
                The first Mapper graph.
            G (MapperGraph):
                The second Mapper graph.
            n (int):
                The interleaving parameter. The default is 1.
        """

        self.n = n

        # --- Check that the ranges are compatible
        if np.abs(F.min_f() - G.min_f()) > n or np.abs(F.max_f() - G.max_f()) > n:
            raise ValueError(
                f"Function values for F and G are too far apart to interleave with n = {n}. Try initializing with a larger n."
            )

        # ---
        # Containers for matrices for later 

        self.B_down_ = {'F':{}, 'G':{}} # boundary matrix
        # self._B_down = None # don't build the boundary matrices unless needed
        self.B_up_ = {'F':{}, 'G':{}} # boundary matrix
        # self._B_up = None # don't build the boundary matrices unless needed
        # self.D_ = {'F':{}, 'G':{}} # distance matrix
        self._D = None # don't build the distance matrices unless needed
        self.I_ = {'F':{}, 'G':{}} # induced maps

        self.val_to_verts = {'F':{}, 'G':{}} # dictionaries from function values to vertices
        self.val_to_edges = {'F':{}, 'G':{}} # dictionaries from function values to edges

        # ----
        # Make F graphs and smoothed versions
        self.F_ = {}
        self.F_["0"] = F
        self.F_["n"], I_0_V, I_0_E = F.smoothing_and_maps(self.n)
        self.F_["2n"], I_n_V, I_n_E = self.F("n").smoothing_and_maps(self.n)

        # Get the dictionaries needed for the induced maps' block structure
        for key in ["0", "n", "2n"]:
            self.val_to_verts["F"][key] = self.F(key).func_to_vertex_dict()
            self.val_to_edges["F"][key] = self.F(key).func_to_edge_dict()

        # Make the induced map from F_0 to F_n
        self.I_["F"]["0"] = {}
        self.I_["F"]["0"]["V"] = LBM(
            map_dict=I_0_V,
            rows_dict=self.val_to_verts["F"]["n"],
            cols_dict=self.val_to_verts["F"]["0"],
        )
        self.I_["F"]["0"]["E"] = LBM(
            map_dict=I_0_E,
            rows_dict=self.val_to_edges["F"]["n"],
            cols_dict=self.val_to_edges["F"]["0"],
        )

        # Make the induced map from F_n to F_2n
        self.I_["F"]["n"] = {}
        self.I_["F"]["n"]["V"] = LBM(
            I_n_V, self.val_to_verts["F"]["2n"], self.val_to_verts["F"]["n"]
        )
        self.I_["F"]["n"]["E"] = LBM(
            I_n_E, self.val_to_edges["F"]["2n"], self.val_to_edges["F"]["n"]
        )

        # ----
        # Now do the same for G
        self.G_ = {}
        self.G_["0"] = G
        self.G_["n"], I_0_V, I_0_E = G.smoothing_and_maps(self.n)
        self.G_["2n"], I_n_V, I_n_E = self.G_["n"].smoothing_and_maps(self.n)

        # Get the dictionaries needed for the induced maps' block structure
        for key in ["0", "n", "2n"]:
            self.val_to_verts["G"][key] = self.G_[key].func_to_vertex_dict()
            self.val_to_edges["G"][key] = self.G_[key].func_to_edge_dict()

        # Make the induced map from G_0 to G_n
        self.I_["G"]["0"] = {}
        self.I_["G"]["0"]["V"] = LBM(
            rows_dict=self.val_to_verts["G"]["n"],
            cols_dict=self.val_to_verts["G"]["0"],
            map_dict=I_0_V,
        )
        self.I_["G"]["0"]["E"] = LBM(
            map_dict=I_0_E,
            rows_dict=self.val_to_edges["G"]["n"],
            cols_dict=self.val_to_edges["G"]["0"],
        )

        self.I_["G"]["0"]["E"] = LBM(
            I_0_E, self.val_to_edges["G"]["n"], self.val_to_edges["G"]["0"]
        )

        # Make the induced map from G_n to G_2n
        self.I_["G"]["n"] = {}
        self.I_["G"]["n"]["V"] = LBM(
            I_n_V, self.val_to_verts["G"]["2n"], self.val_to_verts["G"]["n"]
        )
        self.I_["G"]["n"]["E"] = LBM(
            I_n_E, self.val_to_edges["G"]["2n"], self.val_to_edges["G"]["n"]
        )

        # End making smoothings and induced maps
        # ----
        # ---
        # Build boundary matrices

        for Graph, graph_name in [(self.F, "F"), (self.G, "G")]:
            for key in [
                "0",
                "n",
            ]:  # Note, we don't need to do this for 2n because the matrices are never used.

                B_down = LBM()
                B_up = LBM()

                for i in self.val_to_verts[graph_name][key]:
                    edges = self.val_to_edges[graph_name][key].get(i, [])
                    verts_down = self.val_to_verts[graph_name][key][i]
                    verts_up = self.val_to_verts[graph_name][key].get(i + 1, [])
                    B_down[i] = LM(rows=verts_down, cols=edges)
                    B_up[i] = LM(rows=verts_up, cols=edges)

                    for e in edges:
                        B_down[i][e[0], e] = 1
                        B_up[i][e[1], e] = 1

                min_i = min(list(self.val_to_verts[graph_name][key].keys()))
                max_i = max(list(self.val_to_verts[graph_name][key].keys()))

                min_verts = self.val_to_verts[graph_name][key][min_i]
                max_verts = self.val_to_verts[graph_name][key][max_i]

                B_up[min_i - 1] = LM(rows=min_verts, cols=[])
                B_down[max_i] = LM(rows=max_verts, cols=[])

                self.B_down_[graph_name][key] = B_down
                self.B_up_[graph_name][key] = B_up

        # End boundary matrices
        # ---

        # ---
        # Build the distance matrices
        # We don't actually need these for the ILP, so we won't build them unless needed.
        # If you need them, you can access them using the property self.D_ which will build them if they haven't been built yet.


        # for (metagraph, name) in [ (self.F_,'F'), (self.G_,'G')]:
        #     for key in ['n', '2n']: # Note, we don't need to do this for 0 because the matrices are never used.
        #         self.D_[name][key] = {'V':{}, 'E':{}}

        #         self.D_[name][key]['V'] = metagraph[key].thickening_distance_matrix()
        #         self.D_[name][key]['E'] = metagraph[key].thickening_distance_matrix(obj_type = 'E')

        # End distance matrices

        # ----
        # phi: F -> G^n

        # Initialize the phi matrices. These will have all 0 entries.
        self.phi_ = {"0": {}, "n": {}}
        self.phi_["0"] = {"V": {}, "E": {}}
        self.phi_["0"]["V"] = LBM(
            None,
            self.val_to_verts["G"]["n"],
            self.val_to_verts["F"]["0"],
            random_initialize=initialize_random_maps,
        )

        self.phi_["0"]["E"] = LBM(
            None,
            self.val_to_edges["G"]["n"],
            self.val_to_edges["F"]["0"],
            random_initialize=initialize_random_maps,
        )
        self.phi_["n"] = {"V": {}, "E": {}}
        self.phi_["n"]["V"] = LBM(
            None,
            self.val_to_verts["G"]["2n"],
            self.val_to_verts["F"]["n"],
            random_initialize=initialize_random_maps,
        )

        self.phi_["n"]["E"] = LBM(
            None,
            self.val_to_edges["G"]["2n"],
            self.val_to_edges["F"]["n"],
            random_initialize=initialize_random_maps,
        )

        # End phi
        # ---

        # ----
        # psi: G -> F^n

        # Initialize the psi matrices. These will have all 0 entries.
        self.psi_ = {"0": {}, "n": {}}

        self.psi_["0"] = {"V": {}, "E": {}}

        self.psi_["0"]["V"] = LBM(
            None,
            self.val_to_verts["F"]["n"],
            self.val_to_verts["G"]["0"],
            random_initialize=initialize_random_maps,
        )
        self.psi_["0"]["E"] = LBM(
            None,
            self.val_to_edges["F"]["n"],
            self.val_to_edges["G"]["0"],
            random_initialize=initialize_random_maps,
        )

        self.psi_["n"] = {"V": {}, "E": {}}
        self.psi_["n"]["V"] = LBM(
            None,
            self.val_to_verts["F"]["2n"],
            self.val_to_verts["G"]["n"],
            random_initialize=initialize_random_maps,
        )
        self.psi_["n"]["E"] = LBM(
            None,
            self.val_to_edges["F"]["2n"],
            self.val_to_edges["G"]["n"],
            random_initialize=initialize_random_maps,
        )

        # End psi
        # ---

 
    ### ----------------
    # Properties
    ### ----------------

    # @property
    # def B_down_(self):
    #     if self._B_down is None:
    #         self._B_down = {'F':{}, 'G':{}}
    #         for (Graph, graph_name) in [(self.F_,'F'), (self.G_,'G')]:
    #             for key in ['0', 'n']: # Note, we don't need to do this for 2n because the matrices are never used.

    #                 B_down = LBM()

    #                 for i in self.val_to_verts[graph_name][key]:
    #                     if i in self.val_to_edges[graph_name][key]:
    #                         edges = self.val_to_edges[graph_name][key][i]
    #                         verts_down = self.val_to_verts[graph_name][key][i]
    #                         B_down[i] = LM(rows = verts_down, cols = edges)

    #                         for e in edges:
    #                             B_down[i][e[0],e] = 1

    #                 min_i = min(list(self.val_to_verts[graph_name][key].keys()))
    #                 max_i = max(list(self.val_to_verts[graph_name][key].keys()))

    #                 min_verts = self.val_to_verts[graph_name][key][min_i]
    #                 max_verts = self.val_to_verts[graph_name][key][max_i]

    #                 B_down[min_i-1] = LM(rows = min_verts, cols = [])
    #                 B_down[max_i] = LM(rows = max_verts, cols = [])

    #                 self._B_down[graph_name][key] = B_down
    #                 # self.B_up_[graph_name][key] = B_up
    #     return self._B_down
    
    # @property
    # def B_up_(self):
    #     if self._B_up is None:
    #         self._B_up = {'F':{}, 'G':{}}
    #         for (Graph, graph_name) in [(self.F_,'F'), (self.G_,'G')]:
    #             for key in ['0', 'n']: # Note, we don't need to do this for 2n because the matrices are never used.

    #                 # B_down = LBM()
    #                 B_up = LBM()

    #                 for i in self.val_to_verts[graph_name][key]:
    #                     if i in self.val_to_edges[graph_name][key]:
    #                         edges = self.val_to_edges[graph_name][key][i]
    #                         # verts_down = self.val_to_verts[graph_name][key][i]
    #                         verts_up = self.val_to_verts[graph_name][key][i+1]
    #                         # B_down[i] = LM(rows = verts_down, cols = edges)
    #                         B_up[i] = LM(rows = verts_up, cols = edges)

    #                         for e in edges:
    #                             # B_down[i][e[0],e] = 1
    #                             B_up[i][e[1],e] = 1

    #                 min_i = min(list(self.val_to_verts[graph_name][key].keys()))
    #                 max_i = max(list(self.val_to_verts[graph_name][key].keys()))

    #                 min_verts = self.val_to_verts[graph_name][key][min_i]
    #                 max_verts = self.val_to_verts[graph_name][key][max_i]

    #                 B_up[min_i-1] = LM(rows = min_verts, cols = [])
    #                 B_up[max_i] = LM(rows = max_verts, cols = [])

    #                 # self.B_down_[graph_name][key] = B_down
    #                 self._B_up[graph_name][key] = B_up
    #     return self._B_up

    @property
    def D_(self):
        if self._D is None:
            self._D = {'F':{}, 'G':{}}
            for (metagraph, name) in [ (self.F_,'F'), (self.G_,'G')]:
                for key in ['n', '2n']: # Note, we don't need to do this for 0 because the matrices are never used.
                    self._D[name][key] = {'V':{}, 'E':{}}

                    self._D[name][key]['V'] = metagraph[key].thickening_distance_matrix()
                    self._D[name][key]['E'] = metagraph[key].thickening_distance_matrix(obj_type = 'E')
        return self._D





    ### ----------------
    # Functions for getting stuff out of all the dictionaries
    ### ----------------

    def F(self, key="0"):
        """
        Get the MapperGraph for :math:`F` with key.

        Parameters:
            key (str) :
                The key for the MapperGraph. Either ``'0'``, ``'n'``, or ``'2n'``. Default is ``'0'``.

        Returns:
            MapperGraph :
                The MapperGraph for :math:`F` with key.
        """
        return self.F_[key]

    def G(self, key="0"):
        """
        Get the MapperGraph for :math:`G` with key.

        Parameters:
            key (str) :
                The key for the MapperGraph. Either ``'0'``, ``'n'``, or ``'2n'``. Default is ``'0'``.

        Returns:
            MapperGraph :
                The MapperGraph for :math:`G` with key.
        """
        return self.G_[key]

    def B(self, graph="F", key="0"):
        """
        Get the boundary matrix for a Mapper graph. This is the matrix with entry :math:`B[v,e]` equal to 1 if vertex :math:`v` is an endpoint of edge :math:`e` and 0 otherwise. Also, the boundary matrix is not computed for ``key = '2n'`` because it is not used in the optimization.

        Parameters:
            graph (str) :
                The graph to get the boundary matrix for. Either ``'F'`` or ``'G'``.
            key (str) :
                The key for the boundary matrix. Either ``'0'`` or ``'n'``.

        Returns:
            LabeledBlockMatrix :
                The boundary matrix for the Mapper graph.
        """
        return (
            self.B_down_[graph][key].to_labeled_matrix()
            + self.B_up_[graph][key].to_labeled_matrix()
        )

    def B_down(self, graph="F", key="0"):
        """
        Get the downward boundary matrix for a Mapper graph. This is the matrix with entry :math:`B[v,e]` equal to 1 if vertex :math:`v` is a *lower* endpoint of edge :math:`e` and 0 otherwise. Also, the boundary matrix is not computed for ``key = '2n'`` because it is not used in the optimization.


        Parameters:
            graph (str) :
                The graph to get the boundary matrix for. Either ``'F'`` or ``'G'``.
            key (str) :
                The key for the boundary matrix. Either ``'0'``, or ``'n'``.

        Returns:
            LabeledBlockMatrix :
                The downward boundary matrix for the Mapper graph.
        """
        return self.B_down_[graph][key]

    def B_up(self, graph="F", key="0"):
        """
        Get the upward boundary matrix for a Mapper graph. This is the matrix with entry :math:`B[v,e]` equal to 1 if vertex :math:`v` is an *upper* endpoint of edge :math:`e` and 0 otherwise. Also, the boundary matrix is not computed for ``key = '2n'`` because it is not used in the optimization.

        If ``shift_indices`` is True, the indices of the matrix will be shifted (DOWN?) by one to make matrix multiplication work later.

        Parameters:
            graph (str) :
                The graph to get the boundary matrix for. Either ``'F'`` or ``'G'``.
            key (str) :
                The key for the boundary matrix. Either ``'0'`` or  ``'n'``.

        Returns:
            LabeledBlockMatrix :
                The upward boundary matrix for the Mapper graph.
        """

        return self.B_up_[graph][key]

    def I(self, graph="F", key="0", obj_type="V"):
        """
        Get the induced map from one Mapper graph to another, specifically from ``graph_key`` to ``graph_(key+n)`` sending ``obj_type`` to the same type. For example, ``I('G', 'n', 'E')`` is the map for edges from :math:`G_n` to :math:`G_{2n}`.

        This is the matrix with entry :math:`I[u, v] = 1` if vertex :math:`v` in the first graph maps to vertex :math:`u` in the second graph.

        Parameters:
            graph (str) :
                The graph to get the induced map for. Either ``'F'`` or ``'G'``.
            key (str) :
                The key for the induced map. Either ``'0'`` or ``'n'``.
            obj_type (str) :
                The type of map. Either ``'V'`` or ``'E'``.

        Returns:
            LabeledBlockMatrix :
                The induced map from ``graph_key`` to ``graph_(key+n)``.
        """
        return self.I_[graph][key][obj_type]

    def D(self, graph="F", key="n", obj_type="V"):
        """
        Get the distance matrix for a Mapper graph. This is the matrix with entry :math:`D[u, v]` equal to the minimum thickening needed for vertices :math:`u` and :math:`v` to map to the same connected component (similarly for edges). Note this distance is only defined for vertices or edges at the same function value. Also, the distance matrix is not computed for key = '0' because it is not used in the optimization.

        Parameters:
            graph (str) :
                The graph to get the distance matrix for. Either ``'F'`` or ``'G'``.
            key (str) :
                The key for the distance matrix. Either ``'n'``, or ``'2n'``.

        Returns:
            LabeledBlockMatrix :
                The distance matrix for the Mapper graph.
        """
        return self.D_[graph][key][obj_type]

    def phi(self, key="0", obj_type="V"):
        """
        Get the interleaving map :math:`\\varphi: F \\to G^n` if `key == '0'` or :math:`\\varphi_n: F^n \\to G^{2n}` if `key == 'n'`.

        Parameters:
            key (str) :
                The key for the map. Either ``'0'`` or ``'n'``.
            obj_type (str) :
                The type of map. Either ``'V'`` or ``'E'``.

        Returns:
            LabeledBlockMatrix :
                The interleaving map.
        """
        return self.phi_[key][obj_type]

    def psi(self, key="0", obj_type="V"):
        """
        Get the interleaving map :math:`\\psi: G \\to F^n` if `key == '0'` or :math:`\\psi_n: G^n \\to F^{2n}` if `key == 'n'`.

        Parameters:
            key (str) :
                The key for the map. Either ``'0'`` or ``'n'``.
            obj_type (str) :
                The type of map. Either ``'V'`` or ``'E'``.

        Returns:
            LabeledBlockMatrix :
                The interleaving map.
        """
        return self.psi_[key][obj_type]

    def get_interleaving_map(self, maptype="phi", key="0", obj_type="V"):
        """
        Get the relevant interleaving map. Helpful for iterating over options.

        Parameters:
            maptype (str) :
                The type of map. Either ``'phi'`` or ``'psi'``.
            key (str) :
                The key for the map. Either ``'0'`` or ``'n'``.
            obj_type (str) :
                The type of map. Either ``'V'`` or ``'E'``.

        Returns:
            LabeledBlockMatrix :
                The relevant interleaving map.
        """

        if maptype == "phi":
            return self.phi(key, obj_type)
        elif maptype == "psi":
            return self.psi(key, obj_type)
        else:
            raise ValueError(f"Unknown maptype {maptype}. Must be 'phi' or 'psi'.")

    ### ----------------
    # Functions to set phi and psi matrices for the interleaving (instead of random)
    ### ----------------

    def set_interleaving_maps(self, phi_dict=None, psi_dict=None):
        """
        Set the phi and psi matrices to a given value. Instead of replacing the matrices, set the values block by block.

        Parameters:
            phi_dict (dict):
                A dictionary of the form ``{'0': {'V': phi_0_V, 'E': phi_0_E}, 'n': {'V': phi_n_V, 'E': phi_n_E}}`` where each ``phi_i_j`` is a LabeledBlockMatrix.
            psi_dict (dict):
                A dictionary of the form ``{'0': {'V': psi_0_V, 'E': psi_0_E}, 'n': {'V': psi_n_V, 'E': psi_n_E}}`` where each ``psi_i_j`` is a LabeledBlockMatrix.
        """

        if phi_dict is not None:
            for thickening in ["0", "n"]:
                for obj_type in ["V", "E"]:

                    # fist get the funciton values of the map that's coming in
                    keys = phi_dict[thickening][obj_type].get_all_block_indices()

                    # Now set the values
                    for key in keys:
                        self.phi_[thickening][obj_type][key] = phi_dict[thickening][
                            obj_type
                        ][key]

        if psi_dict is not None:
            for thickening in ["0", "n"]:
                for obj_type in ["V", "E"]:

                    # fist get the funciton values of the map that's coming in
                    keys = psi_dict[thickening][obj_type].get_all_block_indices()

                    # Now set the values
                    for key in keys:
                        self.psi_[thickening][obj_type][key] = psi_dict[thickening][
                            obj_type
                        ][key]

    def set_single_assignment(self, obj_a, obj_b, maptype="phi", key="0", objtype="V"):
        """Set a single assignment in the interleaving map.

        This will be maptype_key(obj_a) = obj_b.

        Note that edges can be passed as a triple (u,v,count) where u and v are the vertices and count is the key for the edge. If passed as a pair, the count will be assumed to be 0.

        Args:
            obj_a (int): Object in the domain of the map.
            obj_b (int): Object in the codomain of the map.
            maptype (str, optional): Map to be set, either 'phi' or 'psi'. Defaults to 'phi'.
            objtype (str, optional): Type of object as input, either 'V' or 'E'. Defaults to 'V'.

        """
        if maptype == "phi":
            start_graph = self.F(key=key)
            if key == "0":
                end_graph = self.G(key="n")
            elif key == "n":
                end_graph = self.G(key="2n")
            else:
                raise ValueError("'key' must be '0' or 'n'.")
        elif maptype == "psi":
            start_graph = self.G(key=key)
            if key == "0":
                end_graph = self.F(key="n")
            elif key == "n":
                end_graph = self.F(key="2n")
            else:
                raise ValueError("'key' must be '0' or 'n'.")
        else:
            raise ValueError("'maptype' must be 'phi' or 'psi'.")

        if objtype == "V":
            # Check that the objects have the same function value
            if start_graph.f[obj_a] != end_graph.f[obj_b]:
                i = start_graph.f[obj_a]
                j = end_graph.f[obj_b]
                end_graph.draw()
                raise ValueError(
                    f"The objects must have the same function value. {obj_a}: {i} != {obj_b}: {j}. Nodes: {start_graph.f.keys()}, {end_graph.f.keys()}"
                )

            i = start_graph.f[obj_a]
        elif objtype == "E":
            if len(obj_a) == 2:
                obj_a = obj_a + (0,)
            elif len(obj_a) != 3:
                raise ValueError("Edges must be passed as pairs or triples.")
            if len(obj_b) == 2:
                obj_b = obj_b + (0,)
            elif len(obj_b) != 3:
                raise ValueError("Edges must be passed as pairs or triples.")

            if start_graph.f[obj_a[0]] != end_graph.f[obj_b[0]]:

                i = start_graph.f[obj_a[0]]
                j = end_graph.f[obj_b[0]]
                raise ValueError(
                    f"The objects must have the same function value. {i} != {j}"
                )
            i = start_graph.f[obj_a[0]]

        else:
            raise ValueError("'objtype' must be 'V' or 'E'.")

        if maptype == "phi":
            row = self.phi_[key][objtype][i].rows.index(obj_b)
            col = self.phi_[key][objtype][i].cols.index(obj_a)
            self.phi_[key][objtype][i].array[:, col] = 0
            self.phi_[key][objtype][i].array[row, col] = 1
        else:  # Maptype = 'psi'
            row = self.psi_[key][objtype][i].rows.index(obj_b)
            col = self.psi_[key][objtype][i].cols.index(obj_a)
            self.psi_[key][objtype][i].array[:, col] = 0
            self.psi_[key][objtype][i].array[row, col] = 1

    def set_random_assignment(self, random_n=True, seed=None):
        """
        Set the phi and psi matrices to random values. No matter what, the maps from F to Gn and G to Fn will be randomly set.  If ``'random_n'`` is True, the maps from Fn to G2n and Gn to F2n will also be randomly set. Otherwise, we will use matrix tricks to figure out the later from the former.

        Note this functions assumes the phi and psi dictionaries were set on initialization. It will overwrite any contents that are there.
        """
        # ----
        # phi: F -> G^n

        # Set the phi matrices. These will have all random entries.
        self.phi_["0"]["V"] = LBM(
            None,
            self.val_to_verts["G"]["n"],
            self.val_to_verts["F"]["0"],
            random_initialize=True,
            seed=seed,
        )

        self.phi_["0"]["E"] = LBM(
            None,
            self.val_to_edges["G"]["n"],
            self.val_to_edges["F"]["0"],
            random_initialize=True,
            seed=seed,
        )

        if random_n:
            self.phi_["n"]["V"] = LBM(
                None,
                self.val_to_verts["G"]["2n"],
                self.val_to_verts["F"]["n"],
                random_initialize=random_n,
                seed=seed,
            )

            self.phi_["n"]["E"] = LBM(
                None,
                self.val_to_edges["G"]["2n"],
                self.val_to_edges["F"]["n"],
                random_initialize=random_n,
                seed=seed,
            )
        else:
            for obj_type in ["V", "E"]:
                self.phi_["n"][obj_type] = (
                    self.I("G", "n", obj_type)
                    @ self.phi("0", obj_type)
                    @ self.I("F", "0", obj_type).T()
                )
                self.phi_["n"][obj_type] = self.phi_["n"][obj_type].to_indicator()

        # End phi
        # ---

        # ----
        # psi: G -> F^n

        # Set the psi matrices. These will have all random entries.
        self.psi_["0"]["V"] = LBM(
            None,
            self.val_to_verts["F"]["n"],
            self.val_to_verts["G"]["0"],
            random_initialize=True,
        )

        self.psi_["0"]["E"] = LBM(
            None,
            self.val_to_edges["F"]["n"],
            self.val_to_edges["G"]["0"],
            random_initialize=True,
        )

        if random_n:
            self.psi_["n"]["V"] = LBM(
                None,
                self.val_to_verts["F"]["2n"],
                self.val_to_verts["G"]["n"],
                random_initialize=random_n,
            )

            self.psi_["n"]["E"] = LBM(
                None,
                self.val_to_edges["F"]["2n"],
                self.val_to_edges["G"]["n"],
                random_initialize=random_n,
            )
        else:
            for obj_type in ["V", "E"]:
                self.psi_["n"][obj_type] = (
                    self.I("F", "n", obj_type)
                    @ self.psi("0", obj_type)
                    @ self.I("G", "0", obj_type).T()
                )
                self.psi_["n"][obj_type] = self.psi_["n"][obj_type].to_indicator()

        # End psi
        # ---

    ### ----------------
    # Functions for drawing stuff
    ### ----------------

    def draw_all_graphs(self, figsize=(15, 10), **kwargs):
        """Draw all the graphs stored in the Interleave object.

        Args:
            figsize (tuple, optional): Sets the size of the figure. Defaults to (15,10).

        Returns:
            tuple: The figure and axes objects.
        """
        fig, axs = plt.subplots(
            2, 3, figsize=figsize, constrained_layout=True, sharey=True
        )

        self.F().draw(ax=axs[0, 0], **kwargs)
        axs[0, 0].set_title(r"$F_0$")

        self.F("n").draw(ax=axs[0, 1], **kwargs)
        axs[0, 1].set_title(r"$F_n$")

        self.F("2n").draw(ax=axs[0, 2], **kwargs)
        axs[0, 2].set_title(r"$F_{2n}$")

        self.G().draw(ax=axs[1, 0], **kwargs)
        axs[1, 0].set_title(r"$G_0$")

        self.G("n").draw(ax=axs[1, 1], **kwargs)
        axs[1, 1].set_title(r"$G_n$")

        self.G("2n").draw(ax=axs[1, 2], **kwargs)
        axs[1, 2].set_title(r"$G_{2n}$")

        # turn on grid for all axes
        for ax in axs.flatten():
            ax.grid(True)

        return fig, axs

    def draw_I(self, graph="F", key="0", obj_type="V", ax=None, **kwargs):
        """
        Draw the induced map from one Mapper graph to another.

        Parameters:
            graph (str) :
                The graph to draw the induced map for. Either ``'F'`` or ``'G'``.
            key (str) :
                The key for the induced map. Either ``'0'`` or ``'n'``.

        Returns:
            matplotlib.axes.Axes
                The axes the matrix was drawn on.
        """
        if ax is None:
            ax = plt.gca()

        self.I(graph, key, obj_type).draw(ax, **kwargs)

        ax.set_xlabel(f"{graph}_{key}")
        if key == "0":
            ax.set_ylabel(f"{graph}_n")
        else:
            ax.set_ylabel(f"{graph}_2n")

        return ax

    def draw_all_I(self, graph="F", figsize=(10, 10), **kwargs):
        """
        Draw all the induced maps.

        Parameters:
            graph (str) :
                The graph to draw the induced maps for. Either ``'F'`` or ``'G'``.
            figsize (tuple) :
                The size of the figure. Default is (10,10).

        Returns:
            tuple: The figure and axes objects.
        """
        fig, axs = plt.subplots(2, 2, figsize=figsize, constrained_layout=True)

        self.draw_I(graph, "0", "V", ax=axs[0, 0], **kwargs)
        title = r"Vertices: $" + graph + r"_0 \to " + graph + r"_{n}$"
        axs[0, 0].set_title(title)

        title = r"Edges: $" + graph + r"_0 \to " + graph + r"_{n}$"
        self.draw_I(graph, "0", "E", ax=axs[1, 0], **kwargs)
        axs[1, 0].set_title(title)

        self.draw_I(graph, "n", "V", ax=axs[0, 1], **kwargs)
        title = r"Vertices: $" + graph + r"_n \to " + graph + r"_{2n}$"
        axs[0, 1].set_title(title)

        title = r"Edges: $" + graph + r"_n \to " + graph + r"_{2n}$"
        self.draw_I(graph, "n", "E", ax=axs[1, 1], **kwargs)
        axs[1, 1].set_title(title)

        return fig, axs

    def draw_B(self, graph="F", key="0", ax=None, **kwargs):
        """
        Draw the boundary matrix for a Mapper graph.

        Parameters:
            graph (str) :
                The graph to draw the boundary matrix for. Either ``'F'`` or ``'G'``.
            key (str) :
                The key for the boundary matrix. Either ``'0'``, ``'n'``, or ``'2n'``.

        Returns:
            matplotlib.axes.Axes
                The axes the matrix was drawn on.
        """
        if ax is None:
            ax = plt.gca()

        self.B(graph, key).draw(ax=ax, **kwargs)
        ax.set_title(f"B({graph}_{key})")
        ax.set_xlabel(f"E({graph}_{key})")
        ax.set_ylabel(f"V({graph}_{key})")

        return ax

    def draw_all_B(self, figsize=(18, 18)):
        """
        Draw all the boundary matrices.

        Parameters:
            figsize (tuple) :
                The size of the figure. Default is (24,18).

        Returns:
            tuple: The figure and axes objects.
        """
        fig, axs = plt.subplots(2, 2, figsize=figsize, constrained_layout=True)
        self.draw_B("F", "0", ax=axs[0, 0])
        axs[0, 0].set_title(r"$B(F_0)$")
        self.draw_B("F", "n", ax=axs[0, 1])
        axs[0, 1].set_title(r"$B(F_n)$")
        # self.draw_B('F', '2n', ax = axs[0, 2])
        # axs[0,2].set_title(r'$B(F_{2n})$')

        self.draw_B("G", "0", ax=axs[1, 0])
        axs[1, 0].set_title(r"$B(G_0)$")
        self.draw_B("G", "n", ax=axs[1, 1])
        axs[1, 1].set_title(r"$B(G_n)$")
        # self.draw_B('G', '2n', ax = axs[1, 2])
        # axs[1,2].set_title(r'$B(G_{2n})$')

        return fig, axs

    def draw_D(
        self, graph="F", key="n", obj_type="V", colorbar=True, ax=None, **kwargs
    ):
        """
        Draw the distance matrix for a Mapper graph.

        Parameters:
            graph (str) :
                The graph to draw the distance matrix for. Either ``'F'`` or ``'G'``.
            key (str) :
                The key for the distance matrix. Either  ``'n'``, or ``'2n'``.
            obj_type (str) :
                The type of matrix. Either ``'V'`` or ``'E'``.
            colorbar (bool) :
                Whether to draw a colorbar. Default is True.
            ax ( matplotlib.axes.Axes) :
                The axes to draw the matrix on. If None, the current axes will be used.
            **kwargs (dict) :
                Additional keyword arguments to pass to the drawing function.

        Returns:
            matplotlib.axes.Axes
                The axes the matrix was drawn on.

        """
        if ax is None:
            ax = plt.gca()

        self.D(graph, key, obj_type).draw(ax=ax, colorbar=colorbar, **kwargs)
        ax.set_xlabel(f"V({graph}_{key})")
        ax.set_ylabel(f"V({graph}_{key})")

        return ax

    def draw_phi(self, key="0", obj_type="V", ax=None, **kwargs):
        """
        Draw the map :math:`\\psi: F \\to G^n`.

        Parameters:
            key (str) :
                The key for the map. Either ``'0'`` or ``'n'``.
            obj_type (str) :
                The type of map. Either ``'V'`` or ``'E'``.

        Returns:
            matplotlib.axes.Axes
                The axes the matrix was drawn on.
        """
        if ax is None:
            ax = plt.gca()

        self.phi(key=key, obj_type=obj_type).draw(ax=ax, **kwargs)

        if key == "0":
            G_key = ""
            F_key = "_n"
        elif key == "n":
            G_key = "_n"
            F_key = "_2n"

        ax.set_ylabel(f"{obj_type}(G{F_key})")
        ax.set_xlabel(f"{obj_type}(F{G_key})")

        return ax

    def draw_psi(self, key="0", obj_type="V", ax=None, **kwargs):
        """
        Draw the map :math:`\\psi: G \\to F^n`.

        Parameters:
            key (str) :
                The key for the map. Either ``'0'`` or ``'n'``.
            obj_type (str) :
                The type of map. Either ``'V'`` or ``'E'``.

        Returns:
            matplotlib.axes.Axes
                The axes the matrix was drawn on.
        """
        if ax is None:
            ax = plt.gca()

        self.psi(key=key, obj_type=obj_type).draw(ax=ax, **kwargs)
        if key == "0":
            F_key = ""
            G_key = "_n"
        elif key == "n":
            F_key = "_n"
            G_key = "_2n"

        ax.set_ylabel(f"{obj_type}(G{G_key})")
        ax.set_xlabel(f"{obj_type}(F{F_key})")

        return ax

    def draw_all_phi(self, figsize=(10, 10), **kwargs):
        """
        Draw all the ``phi`` maps.

        Parameters:
            figsize (tuple) :
                The size of the figure. Default is (10,10).

        Returns:
            tuple: The figure and axes objects.

        """
        fig, axs = plt.subplots(2, 2, figsize=figsize, constrained_layout=True)
        self.draw_phi("0", "V", ax=axs[0, 0], **kwargs)
        axs[0, 0].set_title(r"$\varphi_0^V$")
        self.draw_phi("n", "V", ax=axs[0, 1], **kwargs)
        axs[0, 1].set_title(r"$\varphi_n^V$")

        self.draw_phi("0", "E", ax=axs[1, 0], **kwargs)
        axs[1, 0].set_title(r"$\varphi_0^E$")
        self.draw_phi("n", "E", ax=axs[1, 1], **kwargs)
        axs[1, 1].set_title(r"$\varphi_n^E$")

        return fig, axs

    def draw_all_psi(self, figsize=(10, 10), **kwargs):
        """
        Draw all the ``psi`` maps.

        Parameters:
            figsize (tuple) :
                The size of the figure. Default is (10,10).

        Returns:
            tuple: The figure and axes objects.
        """
        fig, axs = plt.subplots(2, 2, figsize=figsize, constrained_layout=True)
        self.draw_psi("0", "V", ax=axs[0, 0], **kwargs)
        axs[0, 0].set_title(r"$\psi_0^V$")
        self.draw_psi("n", "V", ax=axs[0, 1], **kwargs)
        axs[0, 1].set_title(r"$\psi_n^V$")

        self.draw_psi("0", "E", ax=axs[1, 0], **kwargs)
        axs[1, 0].set_title(r"$\psi_0^E$")
        self.draw_psi("n", "E", ax=axs[1, 1], **kwargs)
        axs[1, 1].set_title(r"$\psi_n^E$")

        return fig, axs

    # =======================
    # Functions for checking commutative diagrams
    # =======================

    def _draw_matrix_mult(self, A, B, C, D, titles=["", "", "", ""], figsize=(12, 3)):
        """
        A drawing function to check the matrix mutliplications for each of the three diagram types.
        """

        fig, axs = plt.subplots(1, 4, figsize=figsize, constrained_layout=True)
        if type(A) == LBM:
            A.draw(ax=axs[0], vmin=-1, vmax=1, filltype="nan")
        else:
            A.draw(ax=axs[0], vmin=-1, vmax=1)
        axs[0].set_title(titles[0])

        if type(B) == LBM:
            B.draw(ax=axs[1], vmin=-1, vmax=1, filltype="nan")
        else:
            B.draw(ax=axs[1], vmin=-1, vmax=1)
        axs[1].set_title(titles[1])

        if type(C) == LBM:
            C.draw(ax=axs[2], vmin=-1, vmax=1, colorbar=True, filltype="nan")
        else:
            C.draw(ax=axs[2], vmin=-1, vmax=1, colorbar=True)
        axs[2].set_title(titles[2])
        D.draw(ax=axs[3], colorbar=True, cmap="PuOr")
        axs[3].set_title(titles[3])

        return fig, axs

    # ---- Parallelogram: Edge-Vertex ----

    def parallelogram_Edge_Vert_matrix(
        self,
        maptype="phi",
        up_or_down="down",
        func_val=None,
        draw=False,
    ):
        """
        Check that the parallelogram for the pair :math:`(S_{\\tau_i}\\subset S_{\\sigma_i})` commutes.
        This is the one that relates the edge maps to the vertex maps. Because a function value has both an up and down version, we need to specify which one we want to check with the ``up_or_down`` parameter.

        If ``func_val`` is not None, we will only check the parallelogram for that function value.

        Parameters:
            maptype (str) :
                The type of map for the relevant diagram. Either ``'phi'`` or ``'psi'``.
            up_or_down (str) :
                Whether to check the up or down version of the parallelogram. Either ``'up'``, ``'down'``, or ``'both'``. Default is ``'down'``.
            func_val (int) :
                The function value to check the parallelogram for. If None, we will check all function values for the full matrix.
            draw (bool) :
                Whether to draw the maps. Default is ``False``.

        Returns:
            LabeledMatrix :
                The matrix that gives the thickening required to make the diagram commute.
        """

        if maptype == "phi":
            start_graph = "F"
            end_graph = "G"
            maptype_latex = r"\varphi"
        elif maptype == "psi":
            start_graph = "G"
            end_graph = "F"
            maptype_latex = r"\psi"

        if up_or_down == "down":
            B = self.B_down
            arrow = "\\downarrow"
            interleaving_map_top = self.get_interleaving_map(maptype, "0", "V")
            interleaving_map_bottom = self.get_interleaving_map(maptype, "0", "E")
        elif up_or_down == "up":
            B = self.B_up
            arrow = "\\uparrow"

            # Shift of the top caused by indexing problem:
            interleaving_map_top = self.get_interleaving_map(
                maptype, "0", "V"
            ).to_shifted_blocks(-1)

            interleaving_map_bottom = self.get_interleaving_map(maptype, "0", "E")

        elif up_or_down == "both":
            interleaving_map_top = self.get_interleaving_map(maptype, "0", "V")
            interleaving_map_bottom = self.get_interleaving_map(maptype, "0", "E")
            B = self.B
            arrow = "{ }"
        else:
            raise ValueError(
                f"Unknown up_or_down {up_or_down}. Must be 'up', 'down', or 'both'."
            )

        if func_val is None:
            Top = interleaving_map_top @ B(start_graph, "0")
            Bottom = B(end_graph, "n") @ interleaving_map_bottom

            Result = Top - Bottom

            if up_or_down == "up":
                # Need to undo the shift fix applied earlier
                Result = Result.to_shifted_blocks(1)

            Result_Dist = self.D(end_graph, 'n', 'V') @ Result
        else: 
            if up_or_down == 'down': # tau_i \to \sigma_i

                Top = self.get_interleaving_map(maptype, '0', 'V')[func_val] @ self.B_down(start_graph, '0')[func_val]
                Bottom = self.B_down(end_graph, 'n')[func_val] @ self.get_interleaving_map(maptype, '0', 'E')[func_val]
                Result = Top - Bottom
                Result_Dist = self.D(end_graph, "n", "V")[func_val] @ Result
            elif up_or_down == "up":  # \tau_i \to \sigma_{i+1}
                Top = (
                    self.get_interleaving_map(maptype, "0", "V")[func_val + 1]
                    @ self.B_up(start_graph, "0")[func_val]
                )
                Bottom = (
                    self.B_up(end_graph, "n")[func_val]
                    @ self.get_interleaving_map(maptype, "0", "E")[func_val]
                )
                Result = Top - Bottom
                Result_Dist = self.D(end_graph, "n", "V")[func_val + 1] @ Result
            else:
                raise ValueError(
                    f"Unknown up_or_down {up_or_down}. Must be 'up' or 'down'."
                )

        if draw:
            titles = ["", "", "", ""]
            titles[0] = f"$M_{maptype_latex}^V \\cdot B_{start_graph}^{arrow}$"
            titles[1] = f"$B_{end_graph}^{arrow} \\cdot M_{maptype_latex}^E$"
            titles[2] = titles[0][:-1] + " - " + titles[1][1:]
            titles[3] = f"$D_{{{end_graph}^{{n}}}}^{{V}} \\cdot  ({titles[2][1:-1]})$"

            fig, axs = self._draw_matrix_mult(
                Top, Bottom, Result, Result_Dist, titles=titles
            )

        return Result_Dist

    def parallelogram_Edge_Vert(self, maptype="phi", up_or_down="down", func_val=None):
        """
        Check that the parallelogram for the pair :math:`(S_{\\tau_i}\\subset S_{\\sigma_i})` commutes, and return the maximum value in the matrix.

        """

        Result = self.parallelogram_Edge_Vert_matrix(maptype, up_or_down, func_val)
        return Result.absmax()

    # ---

    # ---- Parallelogram: Thickening ----

    def parallelogram_matrix(
        self,
        maptype="phi",
        obj_type="V",
        func_val=None,
        draw=False,
    ):
        """
        Get the paralellograms for checking that it's a nat trans.
        These are types 3-6 from Liz's Big List.

        If ``func_val`` is not None, we will only check the parallelogram for that function value.

        Parameters:
            maptype (str) :
                The type of map. Either 'phi' or 'psi'.
            obj_type (str) :
                The type of object. Either ``'V'`` or ``'E'``.
            func_val (int) :
                The function value to check the parallelogram for. If None, we will check all function values for the full matrix.
            draw (bool) :
                Whether to draw the maps. Default is False.

        Returns:
            LabeledMatrix :
                The matrix that gives the thickening required to make the diagram commute.
        """

        if maptype == "phi":
            start_graph = "F"
            end_graph = "G"
            maptype_latex = r"\varphi"
        elif maptype == "psi":
            start_graph = "G"
            end_graph = "F"
            maptype_latex = r"\psi"

        if func_val is None:
            Top = self.get_interleaving_map(maptype, "n", obj_type) @ self.I(
                start_graph, "0", obj_type
            )
            Bottom = self.I(end_graph, "n", obj_type) @ self.get_interleaving_map(
                maptype, "0", obj_type
            )
            Result = Top - Bottom
            # Result= Result.to_labeled_matrix() # To make the matrix labeledmatrix

            Result_Dist = self.D(end_graph, "2n", obj_type) @ Result
        else:
            # Do this for a single input function value
            Top = (
                self.get_interleaving_map(maptype, "n", obj_type)[func_val]
                @ self.I(start_graph, "0", obj_type)[func_val]
            )
            Bottom = (
                self.I(end_graph, "n", obj_type)[func_val]
                @ self.get_interleaving_map(maptype, "0", obj_type)[func_val]
            )
            Result = Top - Bottom
            Result_Dist = self.D(end_graph, "2n", obj_type)[func_val] @ Result

        # --- Drawing--- #
        if draw:
            titles = ["", "", "", ""]
            titles[0] = (
                f"$M_{{{maptype_latex}_n}}^{obj_type} \\cdot I_{start_graph}^{obj_type}$"
            )
            titles[1] = f"$I_{{{end_graph}^n}} \\cdot M_{maptype_latex}^{obj_type}$"
            titles[2] = titles[0][:-1] + " - " + titles[1][1:]
            titles[3] = (
                f"$D_{{{end_graph}^{{2n}}}}^{obj_type} \\cdot ({titles[2][1:-1]})$"
            )

            fig, axs = self._draw_matrix_mult(
                Top, Bottom, Result, Result_Dist, titles=titles
            )

        return Result_Dist

    def parallelogram(self, maptype="phi", obj_type="V", func_val=None):
        """
        Get the loss value for the thickening paralellograms
        """
        Result = self.parallelogram_matrix(maptype, obj_type, func_val).absmax()
        return Result

    # --- Triangles ----

    def triangle_matrix(
        self,
        start_graph="F",
        obj_type="V",
        func_val=None,
        draw=False,
    ):
        """
        Get the triangle for checking that it's an interleaving.

        If ``func_val`` is not None, we will only check the parallelogram for that function value.

        Parameters:
            start_graph (str) :
                The starting graph. Either 'F' or 'G'.
            obj_type (str) :
                The type of object. Either ``'V'`` or ``'E'``.
            func_val (int) :
                The function value to check the parallelogram for. If None, we will check all function values for the full matrix.
            draw (bool) :
                Whether to draw the maps. Default is False.

        Returns:
            LabeledMatrix :
                The matrix that gives the thickenin grequired to make the diagram commute
        """

        if start_graph == "F":
            end_graph = "G"
            map1 = "phi"
            map2 = "psi"
            map1_latex = r"\varphi"
            map2_latex = r"\psi"
        elif start_graph == "G":
            end_graph = "F"
            map1 = "psi"
            map2 = "phi"
            map1_latex = r"\psi"
            map2_latex = r"\varphi"
        else:
            raise ValueError(f"Unknown start_graph {start_graph}. Must be 'F' or 'G'.")

        if func_val is None:
            Top = self.I(start_graph, "n", obj_type) @ self.I(
                start_graph, "0", obj_type
            )
            Bottom = self.get_interleaving_map(
                maptype=map2, key="n", obj_type=obj_type
            ) @ self.get_interleaving_map(maptype=map1, key="0", obj_type=obj_type)
            Result = Top - Bottom
            # Result = Result.to_labeled_matrix() # To make the matrix labeledmatrix

            Result_Dist = self.D(start_graph, "2n", obj_type) @ Result

        else:
            Top = (
                self.I(start_graph, "n", obj_type)[func_val]
                @ self.I(start_graph, "0", obj_type)[func_val]
            )
            Bottom = (
                self.get_interleaving_map(maptype=map2, key="n", obj_type=obj_type)[
                    func_val
                ]
                @ self.get_interleaving_map(maptype=map1, key="0", obj_type=obj_type)[
                    func_val
                ]
            )
            Result = Top - Bottom

            Result_Dist = self.D(start_graph, "2n", obj_type)[func_val] @ Result

        if draw:
            titles = ["", "", "", ""]
            titles[0] = (
                f"$I_{{{start_graph}^n}}^{obj_type} \\cdot I_{start_graph}^{obj_type}$"
            )
            titles[1] = (
                f"$M_{{{map2_latex}^n}}^{obj_type} \\cdot M_{{{map1_latex}}}^{obj_type}$"
            )
            titles[2] = titles[0][:-1] + " - " + titles[1][1:]
            titles[3] = (
                f"$D_{{{start_graph}^{{2n}}}}^{obj_type} \\cdot ({titles[2][1:-1]})$"
            )
            fig, axs = self._draw_matrix_mult(
                Top, Bottom, Result, Result_Dist, titles=titles
            )

        return Result_Dist

    def triangle(self, start_graph="F", obj_type="V", func_val=None):
        """
        Get the loss value for the triangle
        """
        Result = self.triangle_matrix(start_graph, obj_type, func_val).absmax()
        Result = np.ceil(Result / 2)
        return Result

    # --- Loss functions ----

    def loss_table(self):
        """
        Returns a table with the loss for each term in the bound. The actual loss is the maximum of these values, and can be found with the ``loss`` method.

        Returns:
            pd.DataFrame :
                A table with the loss value for each term. The 'Loss' column has the loss value.
        """

        loss_list = []

        # All the edge-vertex parallelogram maps
        for maptype in ["phi", "psi"]:
            for up_or_down in ["up", "down"]:
                result = self.parallelogram_Edge_Vert(
                    maptype=maptype, up_or_down=up_or_down
                )
                loss = result
                loss_list.append([f"Edge-Vertex", maptype, up_or_down, loss])

        # All the parallelogram maps
        for maptype in ["phi", "psi"]:
            for obj_type in ["V", "E"]:
                result = self.parallelogram(maptype=maptype, obj_type=obj_type)
                loss = result
                loss_list.append([f"Thickening", maptype, obj_type, loss])

        # ALl the triangle maps
        for obj_type in ["V", "E"]:
            for start_graph in ["F", "G"]:
                result = self.triangle(start_graph=start_graph, obj_type=obj_type)
                loss = result
                loss_list.append([f"Triangle", obj_type, start_graph, loss])
        loss_table = pd.DataFrame(
            loss_list, columns=["Dgm Type", "Req A", "Req B", "Loss"]
        )
        return loss_table

    def loss(self, verbose=False):
        """
        Computes the loss for the interleaving distance. Specifically, if the loss is :math:`L` and the interleaving class was initiated with :math:`n`, then the interleaving distance is at most :math:`n + L`.

        Returns:
            float :
                The loss value.
        """
        loss_table = self.loss_table()

        loss = loss_table["Loss"].max()

        if verbose:
            print(loss_table)
            print(f"\nLoss: {loss}")
            print(f"Interleaving distance bound: {self.n} + {loss} = {self.n + loss}")
        return loss

    def loss_by_block(self):
        """
        Computes the loss for each block of the interleaving distance.

        Returns:
            dict :
                A dictionary with the loss for each block.
        """

        # This is not at all done yet, consider this a placeholder!
        all_func_vals = set(self.F().get_function_values()) | set(
            self.G().get_function_values()
        )
        all_func_vals = list(all_func_vals)
        all_func_vals.sort()

        loss_dict = {}

        for i in all_func_vals:
            # ====
            # Check the matrices with F(\sigma_i) or G(\sigma_i) in the top left
            # ====
            loss_list = []

            # -- Type 3, 5. Vertex type parallelogram
            for Graph, graph_name, maptype in [
                (self.F, "F", "phi"),
                (self.G, "G", "psi"),
            ]:
                obj_type = "V"
                if i in Graph().get_function_values():
                    loss = self.parallelogram(
                        maptype=maptype, obj_type=obj_type, func_val=i
                    )
                    # loss = result.absmax()
                    # print(f"parallelogram loss for {graph_name} {obj_type} {i} is {loss}")
                    loss_list.append(loss)
                else:
                    # This catch is because the two graphs might have different ranges of function values
                    pass

            # -- Type 7, 9. Vertex type triangle
            for Graph, graph_name in [(self.F, "F"), (self.G, "G")]:
                obj_type = "V"
                if i in Graph().get_function_values():
                    loss = self.triangle(
                        start_graph=graph_name, obj_type=obj_type, func_val=i
                    )
                    # loss = result.absmax()
                    loss_list.append(loss)
                else:
                    # This catch is because the two graphs might have different ranges of function values
                    pass

            # ====
            # Check the matrices with F(\tau_i) or G(\tau_i) in the top left
            # ====

            # -- Type 1, 2. Mixed type parallelogram
            for Graph, graph_name, maptype in [
                (self.F, "F", "phi"),
                (self.G, "G", "psi"),
            ]:
                obj_type = "E"
                edge_vals = Graph().get_function_values()
                edge_vals.pop(edge_vals.index(max(edge_vals)))
                if i in edge_vals:
                    for up_or_down in ["up", "down"]:
                        loss = self.parallelogram_Edge_Vert(
                            maptype=maptype, func_val=i, up_or_down=up_or_down
                        )
                        # loss = result.absmax()
                        # print(f"mixed parallelogram loss for {graph_name} {obj_type} {i} is {loss}")
                        loss_list.append(loss)
                else:
                    # This catch is because the two graphs might have different ranges of function values
                    pass

            # -- Type 4, 6. Edge type parallelogram
            for Graph, graph_name, maptype in [
                (self.F, "F", "phi"),
                (self.G, "G", "psi"),
            ]:
                obj_type = "E"
                edge_vals = Graph().get_function_values()
                edge_vals.pop(edge_vals.index(max(edge_vals)))
                if i in edge_vals:
                    loss = self.parallelogram(
                        maptype=maptype, obj_type=obj_type, func_val=i
                    )
                    # loss = result.absmax()
                    # print(f"parallelogram loss for {graph_name} {obj_type} {i} is {loss}")
                    loss_list.append(loss)
                else:
                    # This catch is because the two graphs might have different ranges of function values
                    pass

            # -- Type 8, 10. Edge type triangle
            for Graph, graph_name in [(self.F, "F"), (self.G, "G")]:
                obj_type = "E"
                edge_vals = Graph().get_function_values()
                edge_vals.pop(edge_vals.index(max(edge_vals)))
                if i in edge_vals:
                    loss = self.triangle(
                        start_graph=graph_name, obj_type=obj_type, func_val=i
                    )
                    # loss = result.absmax()
                    # print(f"triangle loss for {graph_name} {obj_type} {i} is {loss}")
                    loss_list.append(loss)
                else:
                    # This catch is because the two graphs might have different ranges of function values
                    pass

            # Store the max loss for this function value
            loss_dict[i] = max(loss_list)

        # Get max loss over all function values
        # flatten the dictionary
        loss_list = list(loss_dict.values())
        return max(loss_list)


    def all_func_vals(self, map = None):
        """
        Get all the function values that are in the graphs.

        Returns:
            list :
                A list of all the function values.
        """
        if map == 'F':
            return self.F().get_function_values()
        elif map == 'G':
            return self.G().get_function_values()
        else:
            all_func_vals = set(self.F().get_function_values()) | set(self.G().get_function_values()) 
            all_func_vals = list(all_func_vals)
            all_func_vals.sort()

            return all_func_vals
    
    def optimize(self, pulp_solver = None):
        """Uses the ILP to check feasibility of an interleaving at the current ``n``. If feasible, stores the corresponding phi and psi maps, which can be returned using ``self.phi`` and ``self.psi``.
        This function requires the `pulp` package to be installed.

        Parameters:
            pulp_solver (pulp.LpSolver): the solver to use for the ILP optimization. If None, the default solver is used.
        Returns:
<<<<<<< HEAD
            int or None:
                Returns boolean True if an optimal solution was found and False otherwise.
=======
            bool:
                Returns True if an optimal solution was found and False otherwise.
>>>>>>> dcf4c3a66e25255da68ce080152d530e110d09e4
            
        """

        map_dict, prob_status = solve_ilp(self, pulp_solver = pulp_solver)

        if prob_status != 'Optimal':
            return False
            
        self.phi_['0'] = {'V': map_dict['phi_0_V'], 'E': map_dict['phi_0_E']}
        self.phi_['n'] = {'V': map_dict['phi_n_V'], 'E': map_dict['phi_n_E']}
        self.psi_['0'] = {'V': map_dict['psi_0_V'], 'E': map_dict['psi_0_E']}
        self.psi_['n'] = {'V': map_dict['psi_n_V'], 'E': map_dict['psi_n_E']}
        
        
        return True

    def dist_optimize(self, pulp_solver = None):
        """Uses the ILP with distance-matrix constraints and returns the optimized loss value. It also stores the corresponding phi and psi maps, which can be returned using ``self.phi`` and ``self.psi``.
        ## NOTE: This version uses the D matrices in the constraints, which makes it slower. ##
        This function requires the `pulp` package to be installed.
        
        Parameters:
            pulp_solver (pulp.LpSolver): the solver to use for the ILP optimization. If None, the default solver is used.
        Returns:
<<<<<<< HEAD
            int or None:
                Returns the optimized loss value
=======
            float:
                The optimized loss value.

        Raises:
            ValueError:
                If the ILP optimization does not converge.
>>>>>>> dcf4c3a66e25255da68ce080152d530e110d09e4
            
        """

        map_dict, loss_val = solve_ilp_dist(self, pulp_solver = pulp_solver)
            
        self.phi_['0'] = {'V': map_dict['phi_0_V'], 'E': map_dict['phi_0_E']}
        self.phi_['n'] = {'V': map_dict['phi_n_V'], 'E': map_dict['phi_n_E']}
        self.psi_['0'] = {'V': map_dict['psi_0_V'], 'E': map_dict['psi_0_E']}
        self.psi_['n'] = {'V': map_dict['psi_n_V'], 'E': map_dict['psi_n_E']}
        
        
        return loss_val



import unittest
import numpy as np
from cereeberus import MergeTree
from cereeberus.data import ex_mergetrees as ex_mt

class TestMergeTree(unittest.TestCase):
    def check_mt(self, MT):
        # A function to do lots of checks on a given Merge Tree. Just run this thing a ton in any unit test to make sure nothing's screwing up the stuff we think we should be maintaining.
        
        # Check that the node list and the keys for f are the same 
        self.assertEqual(set(MT.nodes), set(MT.f.keys()))

        # Check that the node list and the keys for pos_f are the same 
        self.assertEqual(set(MT.nodes), set(MT.pos_f.keys()))

        # Check that the infinite node is in the list of nodes
        self.assertTrue('v_inf' in MT.nodes)
        self.assertTrue(MT.f['v_inf'] == np.inf)

        finiteNodes = MT.get_finite_nodes()
        self.assertEqual(len(finiteNodes), len(MT.nodes)-1)

        # Check that the pos_list always has the f value in the second entry 
        # except for the infinite node
        for v in finiteNodes:
            self.assertEqual(MT.pos_f[v][1], MT.f[v])
        
        self.assertTrue(np.isfinite(MT.pos_f['v_inf'][1]))

        # Check that all edges are pointing to the higher function value node
        for edge in MT.edges:
            v1, v2 = edge[:2]
            self.assertGreater(MT.f[v2], MT.f[v1])

        # Check that adding a node that creates a loop raises an error. 
        # Since everything has a path to $v_inf$, this should always happen.
        with self.assertRaises(ValueError):
            MT.add_node(0, 'v_inf')

    def test_merge_tree(self):

        MT = ex_mt.randomMergeTree(10)
        self.check_mt(MT)

    def test_merge_labels(self):

        MT = ex_mt.randomMergeTree(10)
        
        # Make sure the LCA funciton works. 
        # Plus the infinite node should never be the LCA of two finite nodes.
        self.assertNotEqual(MT.LCA(0,1), 'v_inf')

        Leaves = MT.get_leaves()

        # Generate the LCA matrix just on the leaf set, which is the default 
        M = MT.LCA_matrix()

        # Check that the LCA matrix is symmetric and of the same size as the leaf set 
        self.assertEqual(M.shape[0], len(Leaves))

        # TODO: Add tests for specifc labeling functions, here I only have for the leaf version



if __name__ == '__main__':
    unittest.main()


# Edited from https://yuminlee2.medium.com/union-find-algorithm-ffa9cd7d2dba


class UnionFind:
    """Union find data structure"""

    def __init__(self, vertices):
        self.parent = {vertex: vertex for vertex in vertices}
        self.size = {vertex: 1 for vertex in vertices}
        self.count = len(vertices)

    # Time: O(logn) | Space: O(1)
    def find(self, node):
        while node != self.parent[node]:
            # path compression
            self.parent[node] = self.parent[self.parent[node]]
            node = self.parent[node]
        return node

    # Time: O(1) | Space: O(1)
    def union(self, node1, node2):
        root1 = self.find(node1)
        root2 = self.find(node2)

        # already in the same set
        if root1 == root2:
            return

        if self.size[root1] > self.size[root2]:
            self.parent[root2] = root1
            self.size[root1] += 1
        else:
            self.parent[root1] = root2
            self.size[root2] += 1

        self.count -= 1

    def components_dict(self):
        """Returns a dictionary mapping component representatives to lists of elements in that component.

        Returns:
            dict: A dictionary where keys are component representatives and values are lists of elements in that component.
        """
        components = {}
        for node in self.parent:
            root = self.find(node)
            if root not in components:
                components[root] = []
            components[root].append(node)
        return components


if __name__ == "__main__":
    edges = [
        [0, 2],
        [1, 4],
        [1, 5],
        [2, 3],
        [2, 7],
        [4, 8],
        [5, 8],
    ]
    numberOfElements = 9

    uf = UnionFind(list(range(9)))

    for node1, node2 in edges:
        uf.union(node1, node2)

    print("number of connected components", uf.count)

# output: ('number of connected components', 3)



import numpy as np


def torus(num=50):
    """Create torus object

    args:
        num (int): number of samples to use when creating torus data

    returns:
        3-element tuple containing

        - **X** : X-coordinates
        - **Y** : Y-coordinates
        - **Z** : Z-coordinates

    """

    U = np.linspace(0, 2 * np.pi, num)
    V = np.linspace(0, 2 * np.pi, num)
    U, V = np.meshgrid(U, V)
    X = (np.cos(V)) * np.cos(U)
    Y = (np.cos(V)) * np.sin(U)
    Z = np.sin(V)
    return X, Y, Z



from cereeberus import ReebGraph, MapperGraph
from cereeberus.data import ex_graphs
from cereeberus.data import ex_reebgraphs as ex_rg


def line(a=0, b=3, seed=None):
    """
    Returns the Mapper graph of a simple line as a MapperGraph class. The endpoints have function value a and b respectively.

    Parameters:
        a, b (int): The function values for the two vertices in increasing order.
        seed (int): Optional. The seed to use for the random number generator, which only controls the layout function.

    Returns:
        MapperGraph: The Mapper graph of the line.


    .. figure:: ../../images/line_mapper.png
        :figwidth: 400px

    """

    MG = ex_rg.line(a, b, seed=seed).to_mapper()
    MG.set_pos_from_f(seed=seed)

    return MG


def torus(a=0, b=1, c=4, d=5, delta=0.1, seed=None):
    """
    Returns the Mapper graph of a simple upright torus as a MapperGraph class.

    Parameters:
        a ,b, c, d (int): The integer function values for the four vertices in increasing order.
        delta (float): Optional. The delta value to use for the Mapper graph.
        seed (int): Optional. The seed to use for the random number generator, which only controls the layout function.

    Returns:
        Mapper: The mapper graph of the torus.


    .. figure:: ../../images/torus_mapper.png
        :figwidth: 400px

    """
    # make sure a, b, c, and d are integers
    if not all(isinstance(x, int) for x in [a, b, c, d]):
        raise ValueError("a, b, c, and d must be integers.")
    MG = ex_rg.torus(a, b, c, d, multigraph=True, seed=seed).to_mapper(delta=delta)
    MG.set_pos_from_f(seed=seed)
    return MG


def dancing_man(delta=0.1, seed=None):
    """
    Returns the Mapper graph of the dancing man as a MapperGraph class.

    Parameters:

        seed (int): Optional. The seed to use for the random number generator, which only controls the layout function.

    Returns:
        MapperGraph: The mapper graph of the dancing man.

    .. figure:: ../../images/dancing_man_mapper.png
        :figwidth: 400px

    """
    return ReebGraph(ex_graphs.dancing_man(), seed=seed).to_mapper(delta=delta)


def juggling_man(delta=0.1, seed=None):
    """
    Returns a modified mapper graph of the juggling man as a MapperGraph class. Some vertex locations were moved to make them integers.

    Parameters:
        seed (int): Optional. The seed to use for the random number generator, which only controls the layout function.

    Returns:
        MapperGraph: The mapper graph of the juggling man.

    .. figure:: ../../images/juggling_man_mapper.png
        :figwidth: 400px

    """
    R = ReebGraph(ex_graphs.juggling_man(), seed=seed)
    R.f[9] = 7
    R.f[8] = 6
    R.f[10] = 5

    return R.to_mapper(delta=delta)


def simple_loops(delta=0.1, seed=None):
    """
    Returns the mapper graph of the simple loops example.

    Parameters:
        seed (int): Optional. The seed to use for the random number generator, which only controls the layout function.

    Returns:
        MapperGraph: The mapper graph of the simple loops example.

    .. figure:: ../../images/simple_loops_mapper.png
        :figwidth: 400px

    """
    return ReebGraph(ex_graphs.simple_loops(), seed=seed).to_mapper()


def simple_loops_unordered(seed=None):
    """
    Returns the mapper graph of the unordered simple loops example.

    Parameters:
        seed (int): Optional. The seed to use for the random number generator, which only controls the layout function.

    Returns:
        MapperGraph: The mapper graph of the simple loops example.

    """
    return ReebGraph(ex_graphs.simple_loops_unordered(), seed=seed).to_mapper()


def interleave_example_A(seed=None):
    """
    Returns the mapper graph of the first example for the interleave function.

    Parameters:
        seed (int): Optional. The seed to use for the random number generator, which only controls the layout function.

    Returns:
        MapperGraph: The mapper graph of the first example for the interleave function.

    .. figure:: ../../images/interleave_example_A_mapper.png
        :figwidth: 400px

    """
    return ex_rg.interleave_example_A(seed=seed).to_mapper()


def interleave_example_B(seed=None):
    """
    Returns the mapper graph of the second example for the interleave function.

    Parameters:
        seed (int): Optional. The seed to use for the random number generator, which only controls the layout function.

    Returns:
        MapperGraph: The mapper graph of the second example for the interleave function.

    .. figure:: ../../images/interleave_example_B_mapper.png
        :figwidth: 400px

    """
    return ex_rg.interleave_example_B(seed=seed).to_mapper()



__all__ = [
    "ex_reebgraphs",
    "ex_mergetrees",
    "ex_mappergraphs",
    "ex_embedgraphs",
    "ex_torus",
]

from .ex_reebgraphs import *
from .ex_mergetrees import *
from .ex_mappergraphs import *
from .ex_embedgraphs import *
from .ex_torus import *



from cereeberus import ReebGraph
from cereeberus.data import ex_graphs


def line(a=0, b=1, seed=None):
    """
    Returns the Reeb graph of a simple line as a ReebGraph class. The endpoints have function value a and b respectively.

    Parameters:
        a, b (int): The function values for the two vertices in increasing order.
        seed (int): Optional. The seed to use for the random number generator, which only controls the layout function.

    Returns:
        ReebGraph: The Reeb graph of the line.


    .. figure:: ../../images/line.png
        :figwidth: 400px

    """

    R = ReebGraph()
    R.add_node("a", a)
    R.add_node("b", b)
    R.add_edge("a", "b")
    R.set_pos_from_f(seed=seed)

    return R


def torus(a=0, b=1, c=4, d=5, multigraph=True, seed=None):
    """
    Returns the Reeb graph of a simple upright torus as a ReebGraph class.

    Parameters:
        a ,b, c, d (int): The function values for the four vertices in increasing order.
        multigraph (bool): Optional. If False, then the loop edges will be subdivided so that the resulting graph doesn't have multiple edges between $b$ and $c$. Default is True.
        seed (int): Optional. The seed to use for the random number generator, which only controls the layout function.


    Returns:
        ReebGraph: The Reeb graph of the torus.

    .. figure:: ../../images/torus.png
        :figwidth: 400px

    """
    R = ReebGraph()
    R.add_node("a", a)
    R.add_node("b", b)
    R.add_node("c", c)
    R.add_node("d", d)
    R.add_edge("a", "b")
    R.add_edge("b", "c")
    R.add_edge("b", "c")
    R.add_edge("c", "d")

    if not multigraph:
        R.subdivide_edge("b", "c", "e", (b + c) / 2)
        R.subdivide_edge("b", "c", "f", (b + c) / 2)

    R.set_pos_from_f(seed=seed)

    return R


def dancing_man(seed=None):
    """
    Returns the Reeb graph of the dancing man as a ReebGraph class.

    Parameters:
        seed (int): Optional. The seed to use for the random number generator, which only controls the layout function.

    Returns:
        ReebGraph: The Reeb graph of the dancing man.

    .. figure:: ../../images/dancing_man.png
        :figwidth: 400px
    """
    return ReebGraph(ex_graphs.dancing_man(), seed=seed)


def juggling_man(seed=None):
    """
    Returns the Reeb graph of the juggling man as a ReebGraph class.

    Parameters:
        seed (int): Optional. The seed to use for the random number generator, which only controls the layout function.

    Returns:
        ReebGraph: The Reeb graph of the juggling man.

    .. figure:: ../../images/juggling_man.png
        :figwidth: 400px
    """
    return ReebGraph(ex_graphs.juggling_man(), seed=seed)


def simple_loops(seed=None):
    """
    Returns the Reeb graph of the simple loops example.

    Parameters:
        seed (int): Optional. The seed to use for the random number generator, which only controls the layout function.

    Returns:
        ReebGraph: The Reeb graph of the simple loops example.

    .. figure:: ../../images/simple_loops.png
        :figwidth: 400px
    """
    return ReebGraph(ex_graphs.simple_loops(), seed=seed)


def simple_loops_unordered(seed=None):
    """
    Returns the Reeb graph of the simple loops example.

    Parameters:
        seed (int): Optional. The seed to use for the random number generator, which only controls the layout function.

    Returns:
        ReebGraph: The Reeb graph of the simple loops example.
    """
    return ReebGraph(ex_graphs.simple_loops_unordered(), seed=seed)


def interleave_example_A(seed=None):
    """
    Returns the Reeb graph of the first example for the interleave function.

    Parameters:
        seed (int): Optional. The seed to use for the random number generator, which only controls the layout function.

    Returns:
        ReebGraph: The Reeb graph of the first example for the interleave function.

    .. figure:: ../../images/interleave_example_A.png
        :figwidth: 400px
    """

    R = dancing_man()
    R.subdivide_edge(6, 7, 8, 3)
    R.add_edge(8, 1)
    R.f = {v: 2 * R.f[v] for v in R.nodes()}
    R.set_pos_from_f(seed=seed)

    return R


def interleave_example_B(seed=None):
    """
    Returns the Reeb graph of the second example for the interleave function.

    Parameters:
        seed (int): Optional. The seed to use for the random number generator, which only controls the layout function.

    Returns:
        ReebGraph: The Reeb graph of the second example for the interleave function.

    .. figure:: ../../images/interleave_example_B.png
        :figwidth: 400px
    """
    R = interleave_example_A()
    R.f[8] = 3
    R.f[5] = 3
    R.remove_edge(3, 1)
    R.set_pos_from_f(seed=seed)
    return R



import unittest
from cereeberus.reeb import embeddedgraph
from cereeberus.data.ex_embedgraphs import exampleEmbeddedGraph
import numpy as np


class TestEmbeddedGraph(unittest.TestCase):
    def test_example_graph(self):
        # Make sure we can build a grpah in the first place
        G = exampleEmbeddedGraph()
        self.assertEqual( len(G.nodes), 6)  # assuming my_function squares its input


    def test_add_node(self):
        # Make sure adding a vertex updates the coordiantes list 
        G = exampleEmbeddedGraph()
        G.add_node('G', 1, 2)
        self.assertEqual( len(G.nodes), 7)
        self.assertEqual( len(G.coordinates), 7)

    def test_add_edge(self):
        # Make sure adding an edge updates the edge list
        G = exampleEmbeddedGraph()
        G.add_edge('A', 'B')
        self.assertEqual( len(G.edges), 6)

    def test_get_coordinates(self):
        # Make sure we can get the coordinates of a vertex
        G = exampleEmbeddedGraph(mean_centered=False)
        coords = G.get_coordinates('A')
        self.assertEqual( coords, (1, 2))

    def test_coords_list(self):
        # Make sure the keys in the coordinates list are the same as the nodes
        G = exampleEmbeddedGraph(mean_centered=False)
        self.assertEqual( len(G.nodes), len(G.coordinates))
        self.assertEqual( set(G.nodes), set(G.coordinates.keys()))

    def test_mean_centered_coordinates(self):
        # Make sure the mean centered coordinates are correct
        G = exampleEmbeddedGraph(mean_centered=False)
        G.set_mean_centered_coordinates()
        x_coords = [x for x, y in G.coordinates.values()]

        self.assertAlmostEqual( np.average(x_coords), 0, places = 1)

    def test_reeb_graph(self):
        # Make sure we can build a reeb graph, and that it will collapse vertices
        # This example will collapse the vertices A and B. Even though F is at the same
        # function value, it shouldn't get collapsed. So we should go down by exactly 
        # one vertex and exactly one edge.
        G = exampleEmbeddedGraph()
        R = R = G.reeb_graph_from_direction(3*np.pi/4)
        self.assertEqual( len(R.nodes), len(G.nodes)-1)
        self.assertEqual( len(R.edges), len(G.edges)-1)

    

if __name__ == '__main__':
    unittest.main()


import unittest

import numpy as np
from cereeberus.data.ex_mappergraphs import line, torus
from cereeberus.distance.labeled_blocks import LabeledBlockMatrix as LBM
from cereeberus.distance.labeled_blocks import LabeledMatrix as LM

from cereeberus import Assignment, Interleave, MapperGraph


class TestInterleaving(unittest.TestCase):
    def test_torus_line(self):
        T = torus(0, 2, 10, 12, delta = 1, seed = 17)
        L = line(0, 12)

        myAssgn = Assignment(T, L, n=1, initialize_random_maps=True)

        # Make sure all the matrices have the right row and column labels 

        #--- B ---# 

        for H, graph in [(myAssgn.F, 'F'), (myAssgn.G, 'G')]: 
            for key in ['0', 'n']:
                mygraph = H(key)

                rows_ = mygraph.sorted_vertices()
                cols_ = mygraph.sorted_edges()

                B_mat = myAssgn.B(graph,key)
                self.assertEqual(rows_, B_mat.rows)
                self.assertEqual(cols_, B_mat.cols)

                # Check all columns sum to 2 
                self.assertTrue(np.all(B_mat.col_sum()[[2]]))
                
        #--- I ---# 

        for H, graph in [(myAssgn.F, 'F'), (myAssgn.G, 'G')]: 
            for key in ['0', 'n']:
                mygraph = H(key)

                if key == '0':
                    nextkey = 'n'
                elif key == 'n':
                    nextkey = '2n'

                mynextgraph = H(nextkey)

                # vertex version 
                myI = myAssgn.I(graph, key, 'V')
                cols_ = mygraph.sorted_vertices()
                rows_ = mynextgraph.sorted_vertices()
                self.assertEqual(cols_, myI.get_all_cols())
                self.assertEqual(rows_, myI.get_all_rows())

                # Check all columns sum to 1
                self.assertTrue(np.all(myI.col_sum()[[1]]))

                # edge version
                myI = myAssgn.I(graph, key, 'E')
                cols_ = mygraph.sorted_edges()
                rows_ = mynextgraph.sorted_edges()
                self.assertEqual(cols_, myI.get_all_cols())
                self.assertEqual(rows_, myI.get_all_rows())

                # Check all columns sum to 1
                self.assertTrue(np.all(myI.col_sum()[[1]]))

        #--- Phi ---# 

        # Checks that the intial setting of Phi is correct

        for key, nextkey in [('0','n'), ('n', '2n')]:
            for obj_type in ['V', 'E']:
                myphi = myAssgn.phi(key, obj_type)
                cols = myAssgn.F(key).sorted_vertices() if obj_type == 'V' else myAssgn.F(key).sorted_edges()
                rows = myAssgn.G(nextkey).sorted_vertices() if obj_type == 'V' else myAssgn.G(nextkey).sorted_edges()
                # print(f'key: {key}, nextkey: {nextkey}, obj_type: {obj_type}')
                self.assertEqual(rows, myphi.get_all_rows())
                self.assertEqual(cols, myphi.get_all_cols())

                # Check all columns sum to 1
                self.assertTrue( np.all(myphi.col_sum()== 1)) 


        #--- Psi ---#

        # Checks that the intial setting of Psi is correct

        for key, nextkey in [('0','n'), ('n', '2n')]:
            for obj_type in ['V', 'E']:
                mypsi = myAssgn.psi(key, obj_type)
                cols = myAssgn.G(key).sorted_vertices() if obj_type == 'V' else myAssgn.G(key).sorted_edges()
                rows = myAssgn.F(nextkey).sorted_vertices() if obj_type == 'V' else myAssgn.F(nextkey).sorted_edges()
                # print(f'key: {key}, nextkey: {nextkey}, obj_type: {obj_type}')
                self.assertEqual(rows, mypsi.get_all_rows())
                self.assertEqual(cols, mypsi.get_all_cols())

                # Check all columns sum to 1
                self.assertTrue( np.all(mypsi.col_sum()== 1))

        
        # Check that the non-random setting of the phi_n and psi_n maps works 
        # If this is true, then those diagrams already commute. 

        myAssgn.set_random_assignment(random_n = False)

        for maptype in ['phi', 'psi']:
            for obj_type in ['V', 'E']:
                P = myAssgn.parallelogram_matrix(maptype, obj_type)
                self.assertTrue(np.all(P.col_sum() == 0))

        # check that the optimize function works and has the same output value as recomputing internally
        result = myAssgn.optimize()
        self.assertFalse(result) 

        newmyAssgn = Assignment(T, L, n=2, initialize_random_maps=True)
        new_result = newmyAssgn.optimize()
        self.assertTrue(new_result)

                
        # Check that the dist_optimize function works and has the same output value as recomputing internally 
        
        loss_out = myAssgn.dist_optimize()
        loss_in = myAssgn.loss()
        self.assertEqual(loss_out, loss_in)

    def test_edgeless_level_boundary_blocks(self):
        """Regression test: levels with vertices but no edges should still have boundary blocks."""
        F = MapperGraph()
        F.add_node("f0", 0)
        F.add_node("f1", 1)
        F.add_node("f2", 2)
        F.add_edge("f0", "f1")

        G = MapperGraph()
        G.add_node("g0", 0)
        G.add_node("g1", 1)
        G.add_node("g2", 2)
        G.add_edge("g0", "g1")

        myAssgn = Assignment(F, G, n=1, initialize_random_maps=True)

        # Level 1 has vertices but no outgoing edges, so boundary blocks should exist with zero columns.
        self.assertIn(1, myAssgn.B_up("F", "0").get_all_block_indices())
        self.assertIn(1, myAssgn.B_down("F", "0").get_all_block_indices())
        self.assertEqual(myAssgn.B_up("F", "0")[1].get_array().shape[1], 0)
        self.assertEqual(myAssgn.B_down("F", "0")[1].get_array().shape[1], 0)

        # This used to raise KeyError when solve_ilp accessed missing boundary blocks.
        myAssgn.optimize()










# =================================
# Construct an example Reeb
# graph of a simple upright torus
# =================================

import networkx as nx


def torus_graph():
    """nx graph input to construct the example Reeb graph of a torus

    Returns:
        graph (networkx graph): reeb graph of a torus

    """

    G = nx.MultiGraph()
    G.add_node(0, pos=(1, 1))
    G.add_node(1, pos=(1, 2))
    G.add_edge(0, 1)

    G.add_node(2, pos=(0.5, 3))
    G.add_node(3, pos=(1.5, 3))
    G.add_edge(1, 2)
    G.add_edge(1, 3)

    G.add_node(4, pos=(1, 4))
    G.add_edge(4, 2)
    G.add_edge(4, 3)

    G.add_node(5, pos=(1, 5))
    G.add_edge(4, 5)

    fx = {0: 1, 1: 2, 2: 3, 3: 3, 4: 4, 5: 5}
    nx.set_node_attributes(G, fx, "fx")

    return G


def reeb_torus_no_fx():
    """
    Reeb graph of a torus with no function values

        Returns:
            graph (networkx graph): reeb graph of a torus

    """

    G = nx.MultiGraph()
    G.add_node(0, pos=(1, 1))
    G.add_node(1, pos=(1, 2))
    G.add_edge(0, 1)

    G.add_node(2, pos=(0.5, 3))
    G.add_node(3, pos=(1.5, 3))
    G.add_edge(1, 2)
    G.add_edge(1, 3)

    G.add_node(4, pos=(1, 4))
    G.add_edge(4, 2)
    G.add_edge(4, 3)

    G.add_node(5, pos=(1, 5))
    G.add_edge(4, 5)

    return G


# =================================
# Construct some of Liz's other
# favorite example Reeb graphs
# =================================


def dancing_man():
    """nx graph input to construct the example dancing man

    Returns:
        graph (networkx graph): reeb graph of dancing man

    """

    G = nx.MultiGraph()
    G.add_node(0, pos=(3, 7))
    G.add_node(1, pos=(3, 6))
    G.add_edge(0, 1)

    G.add_node(2, pos=(1, 5))
    G.add_edge(1, 2)

    G.add_node(3, pos=(5, 5))
    G.add_edge(1, 3)

    G.add_node(4, pos=(7, 6))
    G.add_edge(3, 4)

    G.add_node(5, pos=(0, 4))
    G.add_edge(5, 2)

    G.add_node(6, pos=(3, 4))
    G.add_edge(6, 2)
    G.add_edge(6, 3)

    G.add_node(7, pos=(3, 1))
    G.add_edge(7, 6)

    dd = {0: 1, 1: 2, 2: 2, 3: 1, 4: 1, 5: 0, 6: 1, 7: 0}
    nx.set_node_attributes(G, dd, "down_deg")

    ud = {0: 0, 1: 1, 2: 1, 3: 2, 4: 0, 5: 1, 6: 2, 7: 1}
    nx.set_node_attributes(G, ud, "up_deg")

    fx = {0: 7, 1: 6, 2: 5, 3: 5, 4: 6, 5: 4, 6: 4, 7: 1}
    nx.set_node_attributes(G, fx, "fx")

    return G


def juggling_man():
    """Dancing Man Graph with isolates

    Returns:
        graph (networkx graph): reeb graph of juggling man, which is dancing man with isolates
    """

    G = nx.MultiGraph()
    G.add_node(0, pos=(3, 7))
    G.add_node(1, pos=(3, 6))
    G.add_edge(0, 1)

    G.add_node(2, pos=(1, 5))
    G.add_edge(1, 2)

    G.add_node(3, pos=(5, 5))
    G.add_edge(1, 3)

    G.add_node(4, pos=(7, 6))
    G.add_edge(3, 4)

    G.add_node(5, pos=(0, 4))
    G.add_edge(5, 2)

    G.add_node(6, pos=(3, 4))
    G.add_edge(6, 2)
    G.add_edge(6, 3)

    G.add_node(7, pos=(3, 1))
    G.add_edge(7, 6)

    G.add_node(8, pos=(6.5, 6.5))
    G.add_node(9, pos=(2, 6.5))
    G.add_node(10, pos=(0.75, 5.5))

    dd = {0: 1, 1: 2, 2: 2, 3: 1, 4: 1, 5: 0, 6: 1, 7: 0, 8: 0, 9: 0, 10: 0}
    nx.set_node_attributes(G, dd, "down_deg")

    ud = {0: 0, 1: 1, 2: 1, 3: 2, 4: 0, 5: 1, 6: 2, 7: 1, 8: 0, 9: 0, 10: 0}
    nx.set_node_attributes(G, ud, "up_deg")

    fx = {0: 7, 1: 6, 2: 5, 3: 5, 4: 6, 5: 4, 6: 4, 7: 1, 8: 6.5, 9: 6.5, 10: 5.5}
    nx.set_node_attributes(G, fx, "fx")

    return G


def simple_loops():
    """Simple loops example for plotting loops

    Returns:
        graph (networkx graph): reeb graph with simple loops
    """

    G = nx.MultiGraph()
    G.add_node(0)
    G.add_node(1)
    G.add_node(2)
    G.add_node(3)
    G.add_edge(0, 1)
    G.add_edge(0, 1)
    G.add_edge(1, 2)
    G.add_edge(2, 3)
    G.add_edge(2, 3)

    dd = {0: 0, 1: 2, 2: 1, 3: 2}
    nx.set_node_attributes(G, dd, "down_deg")

    ud = {0: 2, 1: 1, 2: 2, 3: 0}
    nx.set_node_attributes(G, ud, "up_deg")

    fx = {0: 0, 1: 1, 2: 2, 3: 3}
    nx.set_node_attributes(G, fx, "fx")

    return G


def simple_loops_unordered():
    """Simple loops example for plotting loops and testing with unordered edges

    Returns:
        graph (networkx graph): reeb graph with unordered edges
    """

    G = nx.MultiGraph()
    G.add_node(0)
    G.add_node(1)
    G.add_node(2)
    G.add_node(3)
    G.add_edge(0, 1)
    G.add_edge(1, 0)
    G.add_edge(1, 2)
    G.add_edge(2, 3)
    G.add_edge(3, 2)

    dd = {0: 0, 1: 2, 2: 1, 3: 2}
    nx.set_node_attributes(G, dd, "down_deg")

    ud = {0: 2, 1: 1, 2: 2, 3: 0}
    nx.set_node_attributes(G, ud, "up_deg")

    fx = {0: 0, 1: 1, 2: 2, 3: 3}
    nx.set_node_attributes(G, fx, "fx")

    return G



import os, sys


class HiddenPrints:
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


if __name__ == "__main__":
    with HiddenPrints():
        # Code that prints a lot
        print("This will not print")

    print("This will print normally again")
